import os
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any

import openstep_plist
from fontTools.designspaceLib import InstanceDescriptor
from glyphsLib.builder import UFOBuilder

from gftools.utils import open_ufo


@dataclass
class File:
    path: str
    type: str = None

    @property
    def extension(self):
        return self.path.split(".")[-1]

    @property
    def basename(self):
        return os.path.basename(self.path)

    def __hash__(self):
        return hash(id(self))

    def __str__(self):
        return self.path

    def exists(self):
        return os.path.exists(self.path)

    @property
    def is_glyphs(self):
        return self.is_glyphs_file or self.is_glyphspackage

    @property
    def is_ufo(self):
        # ".ufoz" is a zipped UFO; ufoLib2 (via open_ufo) reads it like a UFO.
        return self.extension in ("ufo", "ufoz") or ".ufo.json" in self.path

    @property
    def is_designspace(self):
        return self.extension == "designspace"

    @property
    def is_font_source(self):
        return self.is_glyphs or self.is_ufo or self.is_designspace

    @property
    def is_glyphs_file(self):
        return self.extension == "glyphs"

    @property
    def is_glyphspackage(self):
        return self.extension == "glyphspackage"

    @property
    def glyphs_format(self) -> int:
        if self.is_glyphs_file:
            return int(self.glyphs_plist.get(".formatVersion", 2))
        elif self.is_glyphspackage:
            return int(self.glyphspackage_fontinfo.get(".formatVersion", 2))
        else:
            raise ValueError(
                "File.glyphs_format should not be accessed on non-Glyphs sources"
            )

    @cached_property
    def is_variable(self) -> bool:
        if self.is_designspace:
            return len(self.designspace.sources) > 1
        if self.is_ufo:
            return False
        # Deal with Glyphs sources in their raw plist form to save parsing with
        # glyphsLib, which is far slower
        if self.is_glyphs_file:
            glyphs_fontinfo = self.glyphs_plist
        elif self.is_glyphspackage:
            glyphs_fontinfo = self.glyphspackage_fontinfo
        else:
            raise ValueError(f"unsure how to determine if {self.path} is variable")
        # Fine for Glyphs v2 & v3
        return len(glyphs_fontinfo["fontMaster"]) > 1 or any(
            custom_parameter["name"] == "Virtual Master"
            for custom_parameter in glyphs_fontinfo["customParameters"]
        )

    @cached_property
    def gsfont(self):
        if self.is_glyphs:
            import glyphsLib

            return glyphsLib.load(self.path)
        return None

    @cached_property
    def glyphs_plist(self) -> dict[str, Any]:
        """Grants raw dictly-typed access to a Glyphs to avoid parsing the full
        font with glyphsLib.

        Note that this could be either Glyphs format v2 or v3."""

        assert (
            self.is_glyphs_file
        ), "File.glyphs_plist should not be accessed on non-glyphs single file sources"
        return openstep_plist.load(open(self.path, encoding="utf-8"))

    @cached_property
    def glyphspackage_fontinfo(self) -> dict[str, Any]:
        """Grants raw dictly-typed access to a Glyphspackage's fontinfo.plist to
        avoid parsing the full font with glyphsLib"""

        assert (
            self.is_glyphspackage
        ), "File.glyphspackage_fontinfo should not be accessed on non-glyphspackage sources"
        fontinfo_path = Path(self.path) / "fontinfo.plist"
        return openstep_plist.load(fontinfo_path.open(encoding="utf-8"))

    @cached_property
    def designspace(self):
        if self.is_designspace:
            from fontTools.designspaceLib import DesignSpaceDocument

            return DesignSpaceDocument.fromfile(self.path)
        return None

    @cached_property
    def instances(self):
        if self.is_glyphs:
            gsfont = self.gsfont
            builder = UFOBuilder(gsfont, minimal=True)
            builder.to_designspace_instances()
            return builder._designspace.instances
        elif self.is_designspace:
            return self.designspace.instances
        else:  # UFO
            return [InstanceDescriptor(filename=self.basename)]

    @cached_property
    def family_name(self):
        # Figure out target name
        if self.is_glyphs_file:
            # Optimisation: pull directly from source instead of parsing with
            # glyphsLib
            # Fine for Glyphs v2 & v3
            name = self.glyphs_plist["familyName"]
        elif self.is_glyphspackage:
            # Optimisation: pull this directly from the fontinfo.plist instead
            # of parsing with glyphsLib
            name = self.glyphspackage_fontinfo["familyName"]
        elif self.is_ufo:
            ufo = open_ufo(self.path)
            name = ufo.info.familyName
        elif self.designspace.sources[0].familyName:
            return self.designspace.sources[0].familyName
        else:
            self.designspace.loadSourceFonts(open_ufo)
            return self.designspace.sources[0].font.info.familyName
        return name



from dataclasses import dataclass
import importlib
import inspect
from typing import List
from gftools.builder.file import File

filecache = {}


def get_file(path):
    if path not in filecache:
        filecache[path] = File(path)
    return filecache[path]


@dataclass
class RecipeProviderBase:
    config: dict
    builder: "gftools.builder.GFBuilder"

    def write_recipe(self):
        raise NotImplementedError

    @property
    def sources(self) -> List[File]:
        return [get_file(str(p)) for p in self.config["sources"]]


def get_provider(provider: str):
    # First try gftools.builder.recipeproviders.X
    try:
        mod = importlib.import_module("gftools.builder.recipeproviders." + provider)
    except ModuleNotFoundError:
        # Then try X
        try:
            mod = importlib.import_module(provider)
        except ModuleNotFoundError:
            raise ValueError(f"Cannot find recipe provider {provider}")
    classes = [
        (name, cls)
        for name, cls in inspect.getmembers(mod, inspect.isclass)
        if "RecipeProviderBase" not in name
        and issubclass(cls, RecipeProviderBase)
        and provider in cls.__module__
    ]
    if len(classes) > 1:
        raise ValueError(
            "Multiple recipe providers found in module %s: %s"
            % (provider, [x[0] for x in classes])
        )
    return classes[0][1]



#!/usr/bin/env python3
#
# Copyright 2017 The Fontbakery Authors
# Copyright 2017 The Google Font Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import print_function

import argparse
import os
import sys
import requests
from gfmetadata import FamilyProto
from gfmetadata import text_format

description = (
    "Comparison of category fields of local METADATA.pb files"
    " with data corresponding metadata on the Google Fonts Developer API.\n\n"
    " In order to use it you need to provide an API key."
)
parser = argparse.ArgumentParser(description=description)
parser.add_argument("key", help="Key from Google Fonts Developer API")
parser.add_argument(
    "repo", help=("Directory tree that contains" " directories with METADATA.pb files.")
)
parser.add_argument(
    "--verbose", help="Print additional information", action="store_true"
)


API_URL = "https://www.googleapis.com/webfonts/v1/webfonts?key={}"


def main(args=None):
    args = parser.parse_args(args)
    response = requests.get(API_URL.format(args.key))
    try:
        webfontList = response.json()["items"]
        webfontListFamilyNames = [item["family"] for item in webfontList]
    except (ValueError, KeyError):
        sys.exit(
            "Unable to load and parse" " list of families from Google Web Fonts API."
        )

    for dirpath, dirnames, filenames in os.walk(args.repo):
        metadata_path = os.path.join(dirpath, "METADATA.pb")
        if not os.path.exists(metadata_path):
            continue

        metadata = FamilyProto()
        text_data = open(metadata_path, "rb").read()
        text_format.Merge(text_data, metadata)
        try:
            family = metadata.name
        except KeyError:
            print(
                ('ERROR: "{}" does not contain' " familyname info.").format(
                    metadata_path
                ),
                file=sys.stderr,
            )
            continue

        try:
            index = webfontListFamilyNames.index(family)
            webfontsItem = webfontList[index]
        except ValueError:
            if args.verbose:
                print(
                    (
                        'ERROR: Family "{}" could not be found'
                        " in Google Web Fonts API."
                    ).format(family)
                )
            continue

        if metadata.category == "SANS_SERIF":  # That's fine :-)
            category = "sans-serif"
        else:
            category = metadata.category.lower()

        if category != webfontsItem["category"]:
            print(
                (
                    'ERROR: "{}" category "{}" in git'
                    ' does not match category "{}"'
                    " in API."
                ).format(family, metadata.category, webfontsItem["category"])
            )
        else:
            if args.verbose:
                print(
                    ('OK: "{}" ' 'category "{}" in sync.').format(
                        family, metadata.category
                    )
                )


if __name__ == "__main__":
    main()



from gftools.builder.operations import OperationBase


class Exec(OperationBase):
    description = "Run an arbitrary executable"
    rule = "$exe $args"

    def validate(self):
        if "exe" not in self.original:
            raise ValueError("No executable given")

        if "args" not in self.original:
            raise ValueError("No arguments given")

    @property
    def variables(self):
        # We would like to be able to use the "$in" and
        # "$out" template variables inside our "args" string,
        # but ninja does not perform this second level of
        # variable expansion. So we do it ourselves.
        vars = {
            "exe": self.original["exe"],
            "args": self.original["args"],
        }
        all_input_files = " ".join([source.path for source in self._sources])
        vars["args"] = vars["args"].replace("$in", all_input_files)
        if "$out" in vars["args"]:
            if len(self._targets) != 1:
                raise ValueError("Multiple outputs, but $out used")
            vars["args"] = vars["args"].replace("$out", self.first_target.path)
        return vars



from gftools.builder.operations.fontc import FontcOperationBase


class FontcBuildTTF(FontcOperationBase):
    description = "Build a TTF from a source file (with fontc)"
    rule = "'$fontc_path' -o $out $in $args"



import os
from tempfile import NamedTemporaryFile, TemporaryDirectory

import yaml

from gftools.builder.file import File
from gftools.builder.operations import OperationBase


class AddSubset(OperationBase):
    description = "Add a subset from another font"
    rule = "gftools-add-ds-subsets $args -j -y $yaml -o $out $in"

    def validate(self):
        # Ensure there is a new name
        if not self.first_source.is_font_source:
            raise ValueError("%s is not a font source file" % self.first_source)
        if "subsets" not in self.original:
            raise ValueError("No subsets defined")

    def convert_dependencies(self, graph):
        self._target = TemporaryDirectory()  # Stow object
        self._orig = NamedTemporaryFile(delete=False, mode="w")
        yaml.dump(self.original["subsets"], self._orig)
        self._orig.close()

    @property
    def targets(self):
        if "directory" in self.original:
            target = self.original["directory"]
        else:
            target = self._target.name
        dspath = os.path.join(
            target, self.first_source.basename.rsplit(".", 1)[0] + ".designspace"
        )
        return [File(dspath)]

    @property
    def variables(self):
        return {
            "yaml": self._orig.name,
            "args": self.original.get("args"),
        }



from argparse import ArgumentParser


class GFArgumentParser(ArgumentParser):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_argument(
            "--show-tracebacks", action="store_true", help="Show tracebacks"
        )



#!/usr/bin/env python3
#
# Copyright 2017,2021 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper APIs for interaction with Google Fonts.

Provides APIs to interact with font subsets, codepoints for font or subset.
"""

from __future__ import print_function
from __future__ import unicode_literals

import codecs
import collections
import contextlib
import errno
import os
import re
import sys
import glob
import unicodedata

if __name__ == "__main__":
    # some of the imports here wouldn't work otherwise
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from gfmetadata import FamilyProto
from fontTools import ttLib
from gflanguages import LoadLanguages, parse
from gfmetadata import text_format

# See https://www.microsoft.com/typography/otspec/name.htm.
NAME_COPYRIGHT = 0
NAME_FAMILY = 1
NAME_UNIQUEID = 3
NAME_FULLNAME = 4
NAME_PSNAME = 6


_PLATFORM_ID_MICROSOFT = 3
_PLATFORM_ENC_UNICODE_BMP = 1
_PLATFORM_ENC_UNICODE_UCS4 = 10
_PLATFORM_ENCS_UNICODE = (_PLATFORM_ENC_UNICODE_BMP, _PLATFORM_ENC_UNICODE_UCS4)

_FAMILY_WEIGHT_REGEX = r"([^/-]+)-(\w+)\.[ot]tf$"

# Matches 4 or 5 hexadecimal digits that are uppercase at the beginning of the
# test string. The match is stored in group 0, e.g:
# >>> _NAMELIST_CODEPOINT_REGEX.match('1234X').groups()[0]
# '1234'
# >>> _NAMELIST_CODEPOINT_REGEX.match('1234A').groups()[0]
# '1234A'
_NAMELIST_CODEPOINT_REGEX = re.compile("^([A-F0-9]{4,5})")

# The canonical [to Google Fonts] name comes before any aliases
_KNOWN_WEIGHTS = collections.OrderedDict(
    [
        ("Thin", 100),
        ("Hairline", 100),
        ("ExtraLight", 200),
        ("Light", 300),
        ("Regular", 400),
        ("", 400),  # Family-Italic resolves to this
        ("Medium", 500),
        ("SemiBold", 600),
        ("Bold", 700),
        ("ExtraBold", 800),
        ("Black", 900),
    ]
)

_VALID_STYLES = {"normal", "italic"}

# (Mask, Name) pairs.
# See https://www.microsoft.com/typography/otspec/os2.htm#fss.
_FS_SELECTION_BITS = tuple(
    (1 << i, n)
    for i, n in enumerate(
        (
            "ITALIC",
            "UNDERSCORE",
            "NEGATIVE",
            "OUTLINED",
            "STRIKEOUT",
            "BOLD",
            "REGULAR",
            "USE_TYPO_METRICS",
            "WWS",
            "OBLIQUE",
        )
    )
)


# license_dir => license name mappings
_KNOWN_LICENSE_DIRS = {
    "apache": "APACHE2",
    "ofl": "OFL",
    "ufl": "UFL",
}


FileFamilyStyleWeightTuple = collections.namedtuple(
    "FileFamilyStyleWeightTuple", ["file", "family", "style", "weight"]
)


class Error(Exception):
    """Base for Google Fonts errors."""


class ParseError(Error):
    """Exception used when parse failed."""


def UnicodeCmapTables(font):
    """Find unicode cmap tables in font.

    Args:
      font: A TTFont.
    Yields:
      cmap tables that contain unicode mappings
    """
    for table in font["cmap"].tables:
        if (
            table.platformID == _PLATFORM_ID_MICROSOFT
            and table.platEncID in _PLATFORM_ENCS_UNICODE
        ):
            yield table


def UniqueSort(*args):
    """Returns a sorted list of the unique items from provided iterable(s).

    Args:
      *args: Iterables whose items will be merged, sorted and de-duplicated.
    Returns:
      A list.
    """
    s = set()
    for arg in args:
        s.update(arg)
    return sorted(s)


def RegularWeight(metadata):
    """Finds the filename of the regular (normal/400) font file.

    Args:
      metadata: The metadata to search for the regular file data.
    Returns:
      The name of the regular file, usually Family-Regular.ttf.
    Raises:
      OSError: If regular file could not be found. errno.ENOENT.
    """
    for f in metadata.fonts:
        if f.weight == 400 and f.style == "normal":
            return os.path.splitext(f.filename)[0] + ".ttf"

    name = "??"
    if metadata.HasField("name"):
        name = metadata.name
    raise OSError(errno.ENOENT, "unable to find regular weight in %s" % name)


def Metadata(file_or_dir):
    """Returns fonts_metadata.proto object for a metadata file.

    If file_or_dir is a file named METADATA.pb, load it. If file_or_dir is a
    directory, load the METADATA.pb file in that directory.

    Args:
      file_or_dir: A file or directory.
    Returns:
      Python object loaded from METADATA.pb content.
    Raises:
      ValueError: if file_or_dir isn't a METADATA.pb file or dir containing one.
    """
    if os.path.isfile(file_or_dir) and os.path.basename(file_or_dir) == "METADATA.pb":
        metadata_file = file_or_dir
    elif os.path.isdir(file_or_dir):
        metadata_file = os.path.join(file_or_dir, "METADATA.pb")
        if not os.path.isfile(metadata_file):
            raise ValueError("No METADATA.pb in %s" % file_or_dir)
    else:
        raise ValueError("%s is neither METADATA.pb file or a directory" % file_or_dir)

    msg = FamilyProto()
    with codecs.open(metadata_file, encoding="utf-8") as f:
        text_format.Merge(f.read(), msg)

    return msg


def FamilyName(fontname):
    """Attempts to build family name from font name.

    For example, HPSimplifiedSans => HP Simplified Sans.

    Args:
      fontname: The name of a font.
    Returns:
      The name of the family that should be in this font.
    """
    # SomethingUpper => Something Upper
    fontname = re.sub("(.)([A-Z][a-z]+)", r"\1 \2", fontname)
    # Font3 => Font 3
    fontname = re.sub("([a-z])([0-9]+)", r"\1 \2", fontname)
    # lookHere => look Here
    return re.sub("([a-z0-9])([A-Z])", r"\1 \2", fontname)


def Weight(stylename):
    """Derive weight from a stylename.

    Args:
      stylename: string, e.g. Bold, Regular, or ExtraLightItalic.
    Returns:
      weight: integer
    """
    if stylename.endswith("Italic"):
        return _KNOWN_WEIGHTS[stylename[:-6]]
    return _KNOWN_WEIGHTS[stylename]


def VFWeight(font):
    """Return a variable fonts weight. Return 400 if 400 is within the wght
    axis range else return the value closest to 400

    Args:
      font: TTFont
    Returns:
      weight: integer
    """
    wght_axis = None
    for axis in font["fvar"].axes:
        if axis.axisTag == "wght":
            wght_axis = axis
            break
    value = 400
    if wght_axis:
        if wght_axis.minValue >= 400:
            value = wght_axis.minValue
        if wght_axis.maxValue <= 400:
            value = wght_axis.maxValue
    # TODO (MF) check with GF Eng if we should just assume it's safe to return
    # 400 if a wght axis doesn't exist.
    return int(value)


def Style(stylename):
    return "italic" if "Italic" in stylename else "normal"


def FamilyStyleWeight(path):
    filename = os.path.basename(path)
    if "[" in filename and "]" in filename:
        return VFFamilyStyleWeight(path)
    return FileFamilyStyleWeight(path)


def FileFamilyStyleWeight(path):
    """Extracts family, style, and weight from Google Fonts standard filename.

    Args:
      path: Font path, eg ./fonts/ofl/lobster/Lobster-Regular.ttf.
    Returns:
      FileFamilyStyleWeightTuple for file.
    Raises:
      ParseError: if file can't be parsed.
    """
    m = re.search(_FAMILY_WEIGHT_REGEX, path)
    if not m:
        raise ParseError("Could not parse %s" % path)
    style = Style(m.group(2))
    weight = Weight(m.group(2))
    return FileFamilyStyleWeightTuple(path, FamilyName(m.group(1)), style, weight)


def VFFamilyStyleWeight(path):
    """Extract family, style and weight from a variable font's name table.

    Args:
        path: Font path, eg ./fonts/ofl/lobster/Lobster[wght].ttf.
    Returns:
      FileFamilyStyleWeightTuple for file.
    """
    with ttLib.TTFont(path) as font:
        typoFamilyName = font["name"].getName(16, 3, 1, 1033)
        familyName = font["name"].getName(1, 3, 1, 1033)
        family = (
            typoFamilyName.toUnicode() if typoFamilyName else familyName.toUnicode()
        )

        typoStyleName = font["name"].getName(17, 3, 1, 1033)
        styleName = font["name"].getName(2, 3, 1, 1033)
        style = typoStyleName.toUnicode() if typoStyleName else styleName.toUnicode()
        style = "italic" if "Italic" in style.replace(" ", "") else "normal"
        # For each font in a variable font family, we do not want to return
        # the style's weight. We want to return 400 if 400 is within the
        # the wght axis range. If it isn't, we want the value closest to 400.
        weight = VFWeight(font)
        return FileFamilyStyleWeightTuple(path, family, style, weight)


def ExtractNames(font, name_id):
    return [n.toUnicode() for n in font["name"].names if n.nameID == name_id]


def ExtractName(font_or_file, name_id, default):
    """Extracts a name table field (first value if many) from a font.

    Args:
      font_or_file: path to a font file or a TTFont.
      name_id: the ID of the name desired. Use NAME_* constant.
      default: result if no value is present.
    Returns:
      The value of the first entry for name_id or default if there isn't one.
    """
    value = default
    names = []
    if isinstance(font_or_file, ttLib.TTFont):
        names = ExtractNames(font_or_file, name_id)
    else:
        with contextlib.closing(ttLib.TTFont(font_or_file)) as font:
            names = ExtractNames(font, name_id)

    if names:
        value = names[0]

    return value


def NamePartsForStyleWeight(astyle, aweight):
    """Gives back the parts that go into the name for this style/weight.

    Args:
      astyle: The style name, eg "normal" or "italic"
      aweight: The font weight
    Returns:
      Tuple of parts that go into the name, typically the name for the weight and
      the name for the style, if any ("normal" typically doesn't factor into
      names).
    Raises:
      ValueError: If the astyle or aweight isn't a supported value.
    """
    astyle = astyle.lower()
    if astyle not in _VALID_STYLES:
        raise ValueError("unsupported style %s" % astyle)

    correct_style = None
    if astyle == "italic":
        correct_style = "Italic"

    correct_name = None
    for name, weight in _KNOWN_WEIGHTS.items():
        if weight == aweight:
            correct_name = name
            break

    if not correct_name:
        raise ValueError("unsupported weight: %d" % aweight)

    return tuple([n for n in [correct_name, correct_style] if n])


def _RemoveAll(alist, value):
    while value in alist:
        alist.remove(value)


def FilenameFor(family, style, weight, ext=""):
    family = family.replace(" ", "")
    style_weight = list(NamePartsForStyleWeight(style, weight))
    if "Italic" in style_weight:
        _RemoveAll(style_weight, "Regular")

    style_weight = "".join(style_weight)
    return "%s-%s%s" % (family, style_weight, ext)


def FullnameFor(family, style, weight):
    name_parts = [family]
    name_parts.extend(list(NamePartsForStyleWeight(style, weight)))
    _RemoveAll(name_parts, "Regular")
    return " ".join(name_parts)


def FontDirs(path):
    """Finds all the font directories (based on METADATA.pb) under path.

    Args:
      path: A path to search under.
    Yields:
      Directories under path that have a METADATA.pb.
    """
    for dir_name, _, _ in os.walk(path):
        if os.path.isfile(os.path.join(dir_name, "METADATA.pb")):
            yield dir_name


def FsSelectionMask(flag):
    """Get the mask for a given named bit in fsSelection.

    Args:
      flag: Name of the flag per otspec, eg ITALIC, BOLD, etc.
    Returns:
      Bitmask for that flag.
    Raises:
      ValueError: if flag isn't the name of any fsSelection bit.
    """
    for mask, name in _FS_SELECTION_BITS:
        if name == flag:
            return mask
    raise ValueError("No mask for %s" % flag)


def FsSelectionFlags(fs_selection):
    """Get the named flags enabled in a given fsSelection.

    Args:
      fs_selection: An fsSelection value.
    Returns:
      List of names of flags enabled in fs_selection.
    """
    names = []
    for mask, name in _FS_SELECTION_BITS:
        if fs_selection & mask:
            names.append(name)
    return names


def _EntryForEndOfPath(path, answer_map):
    segments = [s.lower() for s in path.split(os.sep)]
    answers = [answer_map[s] for s in segments if s in answer_map]
    if len(answers) != 1:
        raise ValueError("Found %d possible matches: %s" % (len(answers), answers))
    return answers[0]


def LicenseFromPath(path):
    """Try to figure out the license for a given path.

    Splits path and looks for known license dirs in segments.

    Args:
      path: A filesystem path, hopefully including a license dir.
    Returns:
      The name of the license, eg OFL, UFL, etc.
    Raises:
      ValueError: if 0 or >1 licenses match path.
    """
    return _EntryForEndOfPath(path, _KNOWN_LICENSE_DIRS)


def SupportedLanguages(ttFont, languages=LoadLanguages()):
    """Get languages supported by given ttFont.

    Languages are pulled from the given set. Based on whether exemplar character
    sets are present in the given font.
    """
    chars = [chr(c) for c in ttFont["cmap"].getBestCmap()]
    supported = []
    for lang in languages.values():
        if not lang.HasField("exemplar_chars") or not lang.exemplar_chars.HasField(
            "base"
        ):
            continue
        base = parse(lang.exemplar_chars.base)
        if base.issubset(chars):
            supported.append(lang)
    return supported


def GetExemplarFont(family):
    assert len(family.fonts) > 0, (
        "Unable to select exemplar in family with no fonts: " + family.name
    )
    for font in family.fonts:
        if font.style == "normal" and font.weight == 400:
            # Prefer default style (Regular, not Italic)
            return font
    return family.fonts[0]


def LanguageComments(languages):
    """Generate a mapping for METADATA.pb language field comments.
    Every language field in a METADATA.pb has a comment which is the language's name e.g
      languages: "xh_Latn"  # Xhosa

    Args:
      languages: a dict with keys for lang tags and values for fonts_public_pb2.LanguageProto objects e.g
        languages={"kr_Arab": <class 'fonts_public_pb2.LanguageProto'>}
    Returns:
      A dict with keys for the language field entry and values for the comment.
          {
          'languages: "kr_Arab"': 'Kanuri',
          'languages: "fi_Latn"': 'Finnish'
        }
    """
    line_to_lang_name = {}
    for language in languages.values():
        line = f'languages: "{language.id}"'
        line_to_lang_name[line] = language.name
    return line_to_lang_name


def ReadProto(proto, path):
    with open(path, "r", encoding="utf-8") as f:
        proto = text_format.Parse(f.read(), proto)
        return proto


def WriteProto(proto: FamilyProto, path: str, comments=None):
    with open(path, "w", newline="") as f:
        textproto = text_format.MessageToString(proto, as_utf8=True)
        if comments is not None:
            lines = [
                s if s not in comments else s + "  # " + comments[s]
                for s in textproto.split("\n")
            ]
            textproto = "\n".join(lines)
        f.write(textproto)


def WriteMetadata(proto: FamilyProto, path: str = "METADATA.pb", comments=True):
    if comments is None:
        comment_proto = None
    elif comments is True:
        language = LoadLanguages()
        comment_proto = LanguageComments(language)
    else:
        comment_proto = comments
    WriteProto(proto, path, comment_proto)



#!/usr/bin/env python3
#
# Copyright 2026 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Render strings from a font as a waterfall PNG, or diff two fonts.

Subcommands:
  proof  Render a single font as a waterfall PNG.
  diff   Render before+after waterfalls plus difference image and animated GIF.

Backend defaults to the platform-native rasterizer (CoreText on macOS,
DirectWrite on Windows, FreeType on Linux). Override with ``--backend``.
"""

from __future__ import annotations

import sys
from argparse import ArgumentParser
from pathlib import Path

from gftools.render_text import (
    default_backend,
    diff_image,
    is_variable,
    iter_fvar_instances,
    output_dir_for_all,
    output_dir_for_diff,
    output_path_for,
    output_path_for_instance,
    output_subdir_for_instance,
    pad_to_match,
    parse_variations,
    render_waterfall,
    save_animation,
)


BACKENDS = ("coretext", "directwrite", "freetype")


def main(args=None):
    parser = ArgumentParser(description=__doc__)
    subs = parser.add_subparsers(dest="command", required=True)

    proof = subs.add_parser("proof", help="Render a font as a waterfall PNG.")
    proof.add_argument("font", type=Path, help="Path to a .ttf/.otf font")
    proof.add_argument("text", help="String to render")
    proof.add_argument(
        "-o",
        "--output",
        help=(
            "Output PNG path. With --all, treated as an output directory. "
            "Defaults to <font_stem>.png (or <font_stem>_imgs/ for --all)."
        ),
    )
    group = proof.add_mutually_exclusive_group()
    group.add_argument(
        "--variations",
        help='Variation location, e.g. "wght=400,wdth=75". Default instance if omitted.',
    )
    group.add_argument(
        "--all",
        action="store_true",
        help="Render one image per fvar instance.",
    )
    proof.add_argument(
        "--backend",
        choices=BACKENDS,
        default=None,
        help="Rendering backend. Defaults to the platform-native backend.",
    )
    proof.set_defaults(func=_run_proof)

    diff = subs.add_parser(
        "diff",
        help="Render before+after waterfalls plus a difference image and animated GIF.",
    )
    diff.add_argument("before", type=Path, help="Path to the 'before' font")
    diff.add_argument("after", type=Path, help="Path to the 'after' font")
    diff.add_argument("text", help="String to render")
    diff.add_argument(
        "-o",
        "--output",
        help="Output directory. Default: <after_stem>_diff/ next to the after font.",
    )
    diff_group = diff.add_mutually_exclusive_group()
    diff_group.add_argument(
        "--variations",
        help='Variation location applied to both fonts, e.g. "wght=400".',
    )
    diff_group.add_argument(
        "--all",
        action="store_true",
        help="Produce a diff bundle per fvar instance of the after font.",
    )
    diff.add_argument(
        "--backend",
        choices=BACKENDS,
        default=None,
        help="Rendering backend. Defaults to the platform-native backend.",
    )
    diff.set_defaults(func=_run_diff)

    opts = parser.parse_args(args)
    opts.func(opts)


def _run_proof(opts) -> None:
    backend = opts.backend or default_backend()
    if opts.all:
        _render_all(opts.font, opts.text, opts.output, backend)
        return
    variations = parse_variations(opts.variations) if opts.variations else None
    out = output_path_for(opts.font, variations=variations, output=opts.output)
    img = render_waterfall(opts.font, opts.text, variations=variations, backend=backend)
    img.save(out)
    print(out)


def _render_all(font: Path, text: str, output: str | None, backend: str) -> None:
    if not is_variable(font):
        print(
            f"warning: {font} is a static font — rendering default style only.",
            file=sys.stderr,
        )
        out = output_path_for(font, output=output)
        img = render_waterfall(font, text, backend=backend)
        img.save(out)
        print(out)
        return

    out_dir = output_dir_for_all(font, output_dir=output)
    out_dir.mkdir(parents=True, exist_ok=True)
    for instance_name, location in iter_fvar_instances(font):
        out = output_path_for_instance(font, instance_name, out_dir)
        img = render_waterfall(font, text, variations=location, backend=backend)
        img.save(out)
        print(out)


def _run_diff(opts) -> None:
    backend = opts.backend or default_backend()
    out_dir = output_dir_for_diff(opts.after, output_dir=opts.output)
    out_dir.mkdir(parents=True, exist_ok=True)

    if opts.all and not is_variable(opts.after):
        print(
            f"warning: {opts.after} is a static font — rendering default style only.",
            file=sys.stderr,
        )

    if opts.all and is_variable(opts.after):
        for instance_name, location in iter_fvar_instances(opts.after):
            subdir = output_subdir_for_instance(out_dir, instance_name)
            subdir.mkdir(parents=True, exist_ok=True)
            _emit_diff_bundle(
                opts.before, opts.after, opts.text, location, backend, subdir
            )
    else:
        variations = parse_variations(opts.variations) if opts.variations else None
        _emit_diff_bundle(
            opts.before, opts.after, opts.text, variations, backend, out_dir
        )


def _emit_diff_bundle(
    before_path: Path,
    after_path: Path,
    text: str,
    variations: dict | None,
    backend: str,
    out_dir: Path,
) -> None:
    before_img = render_waterfall(
        before_path, text, variations=variations, backend=backend
    )
    after_img = render_waterfall(
        after_path, text, variations=variations, backend=backend
    )
    before_pad, after_pad = pad_to_match([before_img, after_img])
    diff_img = diff_image(before_pad, after_pad)

    before_pad.save(out_dir / "before.png")
    after_pad.save(out_dir / "after.png")
    diff_img.save(out_dir / "diff.png")
    save_animation(
        [before_pad, after_pad],
        out_dir / "anim.gif",
        labels=["before", "after"],
    )

    for name in ("before.png", "after.png", "diff.png", "anim.gif"):
        print(out_dir / name)


if __name__ == "__main__":
    main()



"""
This script is used to set values in a font file's tables using a YAML-like file.
The configuration file should be formatted as follows:

    OS/2->sTypoAscender: 1200 # Set a value in the OS/2 table
    name->setName: ["Hello world", 0, 3, 1, 0x409] # A method call on the name table
    head->macStyle: |= 0x01  # or with the current value

Unlike standard YAML, this script allows duplicate keys:

    name->setName: ["Hello world", 0, 3, 1, 0x409]
    name->setName: ["Cheese", 1, 3, 1, 0x409]

"""

import argparse
import re
import types

import ruamel.yaml
from fontTools.misc.cliTools import makeOutputFileName
from fontTools.ttLib import TTFont
from ruamel.yaml.constructor import SafeConstructor


# This bit of magic turns the top level dictionary into a list of
# (key, value) pairs, so that we can allow duplicate keys.
# This is helpful to allow us to call the same method (e.g. name->setName)
# multiple times with different arguments.
def construct_yaml_map(self, node):
    if self.deep_construct:
        data = {}
        yield data

        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=True)
            val = self.construct_object(value_node, deep=True)
            data[key] = val
    else:
        data = []
        yield data

        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=True)
            val = self.construct_object(value_node, deep=True)
            data.append((key, val))


SafeConstructor.add_constructor("tag:yaml.org,2002:map", construct_yaml_map)
yaml = ruamel.yaml.YAML(typ="safe")
yaml.allow_duplicate_keys = True


def load_config(fp):
    """
    name->setName: ["Hello world", 0, 3, 1, 0x409]
    OS/2->sTypoAscender: 1200
    -->
    [
        ("name", "setName): ["Hello world", 0, 3, 1, 0x409],
        ("OS/2", "sTypoAscender"): 1200,
    ]
    """
    with open(fp, encoding="utf-8") as doc:
        config = yaml.load(doc)
    return [(tuple(path.split("->")), value) for path, value in config]


def hasmethod(obj, name):
    return hasattr(obj, name) and type(getattr(obj, name)) == types.MethodType


def set_all(obj, config):
    for path, value in config:
        setter(obj, path, value)


def getter(obj, path):
    if len(path) == 0:
        return obj
    key = path[0]
    if hasmethod(obj, key):
        return getattr(obj, key)(*path[1])
    if isinstance(key, str) and hasattr(obj, key):
        return getter(getattr(obj, key), path[1:])
    if isinstance(obj, (list, dict, tuple, TTFont)):
        return getter(obj[key], path[1:])
    return obj


def setter(obj, path, val):
    if len(path) == 0:
        return
    key = path[0]

    if len(path) == 1:
        if isinstance(key, str) and hasmethod(obj, key):
            getattr(obj, key)(*val)
        elif isinstance(key, str) and hasattr(obj, key):
            if isinstance(val, str) and (m := re.match(r"\|=\s*(.*)", val)):
                setattr(obj, key, getattr(obj, key) | int(m.group(1), 0))
            else:
                setattr(obj, key, val)
        elif isinstance(val, str) and (m := re.match(r"\|=\s*(.*)", val)):
            obj[key] = obj[key] | int(m.group(1), 0)
        else:
            obj[path[0]] = val
        return

    if isinstance(key, str) and hasattr(obj, key):
        setter(getattr(obj, key), path[1:], val)
    elif isinstance(obj, (list, dict, tuple, TTFont)):
        is_tuple = False
        # convert numeric keys if needed
        try:
            obj[key]
        except:
            key = str(key)
            obj[key]
        # convert tuples to lists and then reconvert back using backtracking
        if isinstance(obj[key], tuple):
            is_tuple = True
            obj[key] = list(obj[key])
        setter(obj[key], path[1:], val)
        if is_tuple:
            obj[key] = tuple(obj[key])


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("font", type=TTFont)
    parser.add_argument("config")
    out = parser.add_mutually_exclusive_group()
    out.add_argument("--inplace", action="store_true", default=False)
    out.add_argument("--out", "-o", default=None)
    args = parser.parse_args(args)

    config = load_config(args.config)
    set_all(args.font, config)

    if args.inplace:
        args.out = args.font.reader.file.name
    elif not args.out:
        args.out = makeOutputFileName(args.font.reader.file.name)
    args.font.save(args.out)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
gftools gen-stat

Generate a STAT table for each font in a variable font family
using the GF axis registry. Alternatively, users can generate
STAT tables from a yaml file which has the following structure:

```
Lora[wght].ttf:
- name: Weight
  tag: wght
  values:
  - name: Regular
    value: 400
    ...
- name: Width
  tag: wdth
  values:
  ...

Lora-Italic[wght].ttf
...
```

Usage:

# Standard usage. Fonts will have ".fix" appended to their filenames
gftools gen-stat font1.ttf --axis-order wdth wght

# Output fonts to a dir
gftools gen-stat font1.ttf font2.ttf --axis-order wdth wght --out ~/Desktop/out

# Overwrite input fonts
gftools gen-stat font1.ttf font2.ttf --axis-order wdth wght --inplace

# Overide which axis values are elided
gftools gen-stat font.ttf --elided-values wght=400 --axis-order wdth wght

# Generate stats from a file
gftools gen-stat font.ttf --src my_stat.yaml

"""
from axisregistry import AxisRegistry
from fontTools.ttLib import TTFont
from gftools.stat import gen_stat_tables, gen_stat_tables_from_config
import argparse
import yaml
import os

axis_registry = AxisRegistry()


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "fonts", nargs="+", help="Variable TTF files which make up a family"
    )
    parser.add_argument("--src", help="use yaml file to build STAT", default=None)
    parser.add_argument("--out", "-o", help="Output dir for fonts")
    parser.add_argument(
        "--inplace", action="store_true", default=False, help="Overwrite input files"
    )
    args = parser.parse_args(args)

    print(args.fonts)
    fonts = [TTFont(f) for f in args.fonts]

    if args.src:
        config = yaml.load(open(args.src), Loader=yaml.SafeLoader)
        gen_stat_tables_from_config(config, fonts)
    else:
        gen_stat_tables(fonts)

    if args.out:
        if not os.path.isdir(args.out):
            os.mkdir(args.out)

    for font in fonts:
        if args.out:
            dst = os.path.join(args.out, os.path.basename(font.reader.file.name))
        elif args.inplace:
            dst = font.reader.file.name
        else:
            dst = font.reader.file.name + ".fix"
        if os.path.isfile(dst):
            os.remove(dst)
        print(f"Saving font to {dst}")
        font.save(dst)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility to dump codepoints in a font.

Prints codepoints supported by the font, one per line, in hex (0xXXXX).

"""

import os
import sys
import unicodedata
import argparse

from gfsubsets import CodepointsInFont, SubsetsForCodepoint


parser = argparse.ArgumentParser(description="Dump codepoints in a font")
parser.add_argument(
    "--show_char", action="store_true", help="Print the actual character"
)
parser.add_argument(
    "--show_subsets", action="store_true", help="Print what subsets, if any, char is in"
)
parser.add_argument("font", metavar="TTF", nargs="+", help="font files")


def main(args=None):
    args = parser.parse_args(args)

    cps = set()
    for filename in args.font:
        if not os.path.isfile(filename):
            sys.exit("%s is not a file" % filename)
        cps |= CodepointsInFont(filename)

    for cp in sorted(cps):
        show_char = ""
        if args.show_char:
            show_char = " " + chr(cp).strip() + " " + unicodedata.name(chr(cp), "")
        show_subset = ""
        if args.show_subsets:
            show_subset = " subset:%s" % ",".join(SubsetsForCodepoint(cp))

        print("0x%04X%s%s" % (cp, show_char, show_subset))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2017 The Fontbakery Authors
# Copyright 2017 The Google Font Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Check the version number of a family hosted on fonts.google.com.
"""
from __future__ import print_function
from argparse import ArgumentParser
from fontTools.ttLib import TTFont
from ntpath import basename
from zipfile import ZipFile
from gftools.utils import (
    download_family_from_Google_Fonts,
    download_file,
    fonts_from_zip,
)


def parse_version_head(fonts):
    """Return a family's version number. Ideally, each font in the
    family should have the same version number. If not, return the highest
    version number."""
    versions = []
    if isinstance(fonts, list):
        for font in fonts:
            versions.append(float(font["head"].fontRevision))
    else:
        versions.append(float(fonts["head"].fontRevision))
    return max(versions)


def main(args=None):
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("family", help="Name of font family")
    parser.add_argument(
        "-wc", "--web-compare", help="Compare against a web url .zip family"
    )
    parser.add_argument(
        "-lc", "--local-compare", nargs="+", help="Compare against a set of local ttfs"
    )
    args = parser.parse_args(args)

    google_family = download_family_from_Google_Fonts(args.family)
    google_family_fonts = [TTFont(f) for f in google_family]
    google_family_version = parse_version_head(google_family_fonts)

    if args.web_compare:
        if args.web_compare.endswith(".zip"):
            web_family_zip = ZipFile(download_file(args.web_compare))
            web_family = fonts_from_zip(web_family_zip)
            web_family_fonts = [
                TTFont(f) for f in web_family if f.name.endswith(".ttf")
            ]
            web_family_name = set(f.reader.file.name.split("-")[0] for f in web_family)
            web_family_version = parse_version_head(web_family_fonts)
        print(
            "Google Fonts Version of %s is v%s" % (args.family, google_family_version)
        )
        print(
            "Web Version of %s is v%s"
            % (", ".join(web_family_name), web_family_version)
        )

    elif args.local_compare:
        local_family = [TTFont(f) for f in args.local_compare]
        local_family_version = parse_version_head(local_family)
        local_fonts_name = set(basename(f.split("-")[0]) for f in args.local_compare)
        print(
            "Google Fonts Version of %s is v%s" % (args.family, google_family_version)
        )
        print(
            "Local Version of %s is v%s"
            % (",".join(local_fonts_name), local_family_version)
        )

    else:
        print(
            "Google Fonts Version of %s is v%s" % (args.family, google_family_version)
        )


if __name__ == "__main__":
    main()



from gftools.builder.operations import OperationBase


class Avar2ToAvar1(OperationBase):
    description = "Flatten an avar2 variable font into an avar1 variable font"
    rule = "gftools-avar2-to-avar1 $args -o $out $in"



from __future__ import annotations
import json
import logging
import os
import re
from datetime import datetime
from functools import lru_cache

try:
    import tomllib
except ImportError:
    import tomli as tomllib

import requests  # type: ignore
from gftools.push.items import (
    Axis,
    AxisFallback,
    Designer,
    Family,
    FamilyMeta,
    Itemer,
    Items,
)
from gftools.utils import (
    PROD_FAMILY_DOWNLOAD,
)

log = logging.getLogger("gftools.push")


# This module uses api endpoints which shouldn't be public. Ask
# Marc Foley for the .gf_push_config.toml file. Place this file in your
# home directory. Environment variables can also be used instead.
#
# Some servers require basic authentication. Credentials can be provided
# with the SANDBOX_USER/SANDBOX_PASSWORD (and equivalent) environment
# variables, or with a [credentials] table in the push config file:
#
# [credentials]
# sandbox_user = "..."
# sandbox_password = "..."
config_ini = os.path.join(os.path.expanduser("~"), ".gf_push_config.ini")
config_toml = os.path.join(os.path.expanduser("~"), ".gf_push_config.toml")

if os.path.exists(config_ini) and not os.path.exists(config_toml):
    with open(config_ini, "r") as fr:
        with open(config_toml, "w") as fw:
            for line in fr:
                # convert ini format to toml format
                if "=" in line:
                    key, value = line.split("=", 1)
                    fw.write(f'{key.strip()} = "{value.strip()}"\n')
                else:
                    fw.write(line)

if os.path.exists(config_toml):
    with open(config_toml, "rb") as f:
        config = tomllib.load(f)
else:
    config = {
        "urls": {
            "sandbox_family_download": os.environ.get("SANDBOX_FAMILY_DOWNLOAD"),
            "sandbox_meta": os.environ.get("SANDBOX_META_URL"),
            "sandbox_versions": os.environ.get("SANDBOX_VERSIONS_URL"),
            "production_meta": os.environ.get("PRODUCTION_META_URL"),
            "production_versions": os.environ.get("PRODUCTION_VERSIONS_URL"),
        }
    }
SANDBOX_FAMILY_DOWNLOAD = config["urls"]["sandbox_family_download"]
SANDBOX_META_URL = config["urls"]["sandbox_meta"]
SANDBOX_VERSIONS_URL = config["urls"]["sandbox_versions"]
PRODUCTION_META_URL = config["urls"]["production_meta"]
PRODUCTION_VERSIONS_URL = config["urls"]["production_versions"]


def server_auth(server_name: str) -> "tuple[str, str] | None":
    """Basic auth credentials for a server, taken from the push config's
    [credentials] table or the <SERVER>_USER/<SERVER>_PASSWORD environment
    variables e.g SANDBOX_USER, SANDBOX_PASSWORD."""
    creds = config.get("credentials", {})
    user = creds.get(f"{server_name}_user") or os.environ.get(
        f"{server_name.upper()}_USER"
    )
    password = creds.get(f"{server_name}_password") or os.environ.get(
        f"{server_name.upper()}_PASSWORD"
    )
    if user and password:
        return (user, password)
    return None


@lru_cache
def gf_server_metadata(url: str, auth: "tuple[str, str] | None" = None):
    """Get family json data from a Google Fonts metadata url"""
    # can't do requests.get("url").json() since request text starts with ")]}'"
    info = requests.get(url, auth=auth).json()
    return {i["family"]: i for i in info["familyMetadataList"]}


@lru_cache
def gf_server_family_metadata(
    url: str, family: str, auth: "tuple[str, str] | None" = None
):
    """Get metadata for family on a server"""
    # can't do requests.get("url").json() since request text starts with ")]}'"
    url = url + f"/{family.replace(' ', '%20')}"
    r = requests.get(url, auth=auth)
    if r.status_code != 200:
        return None
    text = r.text
    info = json.loads(text[4:])
    return info


class GFServer(Itemer):
    def __init__(
        self,
        name: str,
        url: str = PRODUCTION_META_URL,
        dl_url: str = PROD_FAMILY_DOWNLOAD,
        version_url: str = PRODUCTION_VERSIONS_URL,
        auth: "tuple[str, str] | None" = None,
    ):
        self.name = name
        self.url = url
        self.dl_url = dl_url
        self.version_url = version_url
        # underscore attribs are skipped by Itemer.to_json so credentials
        # never end up in saved server data
        self._auth = auth
        self.families: dict[str, Family] = {}
        self.designers: dict[str, Designer] = {}
        self.metadata: dict[str, FamilyMeta] = {}
        self.axisregistry: dict[str, Axis] = {}
        resp = requests.get(self.version_url, auth=self._auth)
        if resp.status_code != 200:
            raise ValueError(
                f"Failed to get font family versions from the {self.name} "
                f"server (HTTP {resp.status_code}). The server may require "
                "basic authentication. Provide credentials with the "
                f"{self.name.upper()}_USER and {self.name.upper()}_PASSWORD "
                "environment variables or a [credentials] table in the "
                "push config file."
            )
        self.family_versions_data = json.loads(resp.text[5:])
        self.family_versions = {
            i["name"]: i["fontVersions"][0]["version"]
            for i in self.family_versions_data["familyVersions"]
        }

    def is_online(self):
        req = requests.head(self.url, auth=self._auth)
        if req.status_code == 200:
            return True
        return False

    @property
    def last_push(self):
        return datetime.fromtimestamp(
            self.family_versions_data["lastUpdate"]["seconds"]
        )

    def compare_push_item(self, item: Items):
        server_item = self.find_item(item)
        return server_item == item

    def find_item(self, item):
        if isinstance(item, Family):
            server_item = self.families.get(item.name)
        elif isinstance(item, Designer):
            server_item = self.designers.get(item.name)
        elif isinstance(item, Axis):
            server_item = self.axisregistry.get(item.tag)
        elif isinstance(item, FamilyMeta):
            server_item = self.metadata.get(item.name)
        else:
            return None
        return server_item

    def update_axis_registry(self, axis_data):
        for axis in axis_data:
            self.axisregistry[axis["tag"]] = Axis.from_gf_json(axis)

    def update_family(self, name: str):
        family_version = self.family_versions.get(name)
        if family_version:
            self.families[name] = Family(name, family_version)
            return True
        else:
            family = Family.from_gf(name, dl_url=self.dl_url, auth=self._auth)
            if family:
                self.families[name] = family
                return True
        return False

    def update_family_designers(self, name: str):
        meta = gf_server_family_metadata(self.url, name, auth=self._auth)
        for designer in meta["designers"]:
            self.designers[designer["name"]] = Designer.from_gf_json(designer)

    def update_metadata(self, name: str):
        meta = gf_server_family_metadata(self.url, name, auth=self._auth)
        self.metadata[meta["family"]] = FamilyMeta.from_gf_json(meta)

    def update_all(self, last_checked: str):
        resp = requests.get(self.url, auth=self._auth)
        resp.raise_for_status()
        meta = resp.json()
        self.update_axis_registry(meta["axisRegistry"])

        families_data = meta["familyMetadataList"]
        for family_data in families_data:
            family_name = family_data["family"]
            # I beleive this is a test family so we'll skip it for now
            if family_name == "Roboto_old":
                continue
            last_modified = family_data["lastModified"]

            cached_family_version = self.family_versions.get(family_name)
            existing_family_version = self.families.get(family_name)
            # always ensure we repull family data if the family_version api
            # is different
            if cached_family_version and existing_family_version:
                if cached_family_version != existing_family_version.version:
                    self.update(family_name)
            elif last_modified >= last_checked:
                self.update(family_name)

    def update(self, family_name):
        log.info(f"Updating {family_name}")
        if self.update_family(family_name):
            self.update_family_designers(family_name)
            self.update_metadata(family_name)


class GFServers(Itemer):
    SANDBOX = "sandbox"
    PRODUCTION = "production"
    SERVERS = (SANDBOX, PRODUCTION)

    def __init__(self):
        self.last_checked = datetime.fromordinal(1).isoformat().split("T")[0]
        self.sandbox = GFServer(
            GFServers.SANDBOX,
            SANDBOX_META_URL,
            SANDBOX_FAMILY_DOWNLOAD,
            SANDBOX_VERSIONS_URL,
            auth=server_auth(GFServers.SANDBOX),
        )
        self.production = GFServer(
            GFServers.PRODUCTION,
            PRODUCTION_META_URL,
            PROD_FAMILY_DOWNLOAD,
            PRODUCTION_VERSIONS_URL,
            auth=server_auth(GFServers.PRODUCTION),
        )
        self.fp = None

    def last_pushes(self):
        log.info(
            "Last pushes for each server:\n"
            f"Sandbox: {self.sandbox.last_push}\n"
            f"Production: {self.production.last_push}\n"
        )

    def servers_online(self):
        for server in self:
            is_online = server.is_online()
            if not is_online:
                raise ValueError(f"Server {server.name} is offline")

    def __iter__(self):
        for server in GFServers.SERVERS:
            yield getattr(self, server)

    def update_all(self):
        for server in self:
            server.update_all(self.last_checked)
        self.last_checked = datetime.now().isoformat().split("T")[0]

    def update(self, family_name):
        for server in self:
            try:
                server.update(family_name)
            except Exception as e:
                log.error(f"Error updating {family_name} on {server.name}: {e}")

    def compare_item(self, item: Items):
        res = item.to_json()
        for server in self:
            res[f"In {server.name}"] = server.compare_push_item(item)
        return res

    def save(self, fp: "str | Path"):
        from copy import deepcopy

        cp = deepcopy(self)
        # do not save family_versions data. We want to request this each time
        for attr in ["family_versions", "family_versions_data"]:
            for server_name in ["sandbox", "production"]:
                server = getattr(cp, server_name)
                if hasattr(server, attr):
                    delattr(server, attr)
        data = cp.to_json()
        json.dump(data, open(fp, "w", encoding="utf8"), indent=4)

    @classmethod
    def open(cls, fp: "str | Path"):
        data = json.load(open(fp, encoding="utf8"))
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data):
        inst = cls()
        inst.last_checked = data["last_checked"]
        for server_name in GFServers.SERVERS:
            server = getattr(inst, server_name)

            for item_type, item_value in data[server_name].items():
                # if family_versions data is saved, skip it so we get requested
                # data instead
                if item_type in ["family_versions", "traffic_jam"]:
                    continue
                if item_type == "families":
                    server.families = {k: Family(**v) for k, v in item_value.items()}
                elif item_type == "designers":
                    server.designers = {k: Designer(**v) for k, v in item_value.items()}
                elif item_type == "metadata":
                    server.metadata = {
                        k: FamilyMeta(**v) for k, v in item_value.items()
                    }
                elif item_type == "axisregistry":
                    server.axisregistry = {k: Axis(**v) for k, v in item_value.items()}
                    for k, v in server.axisregistry.items():
                        server.axisregistry[k].fallback = [
                            AxisFallback(**a) for a in v.fallback
                        ]
                else:
                    setattr(server, item_type, item_value)
        return inst



"""DirectWrite rendering backend (Windows), implemented via Skia-Python.

We *don't* call DirectWrite directly through comtypes. The rationale:

- DirectWrite is a deep COM hierarchy. A direct comtypes implementation
  needs ~400 lines of vtable scaffolding (interface declarations for
  IDWriteFactory, IDWriteTextAnalyzer, IDWriteGdiInterop,
  IDWriteBitmapRenderTarget, plus DWRITE structs and GDI plumbing for
  pixel readback) before any actual rendering happens.
- Skia delegates to ``SkFontMgr_DirectWrite`` and ``SkScalerContext_DW``
  on Windows, so font loading and glyph rasterisation still hit the
  platform-native DirectWrite stack — Skia just adds a thin
  gamma/contrast layer in front of it. The platform-rendering signal we
  care about (ClearType, subpixel positioning, DWrite hinting) is
  preserved.
- Symmetrically, Skia uses CoreText on macOS and FreeType+FontConfig on
  Linux, so anyone wanting a uniformly Skia-based matrix could use this
  backend on all three platforms via ``--backend directwrite`` (though
  ``coretext`` and ``freetype`` remain the platform defaults).

Fidelity tradeoff: Skia's default text path uses HarfBuzz for shaping,
not DirectWrite's shaper. For Latin pangrams this is invisible; for a
complex-script regression test it would matter and we'd need to
revisit.
"""

from __future__ import annotations

from pathlib import Path

from PIL import Image

try:
    import skia
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "gftools was installed without the QA dependencies. To install the "
        "dependencies, see the ReadMe, "
        "https://github.com/googlefonts/gftools#installation"
    )

PADDING = 4


def render_row(
    font_path: Path,
    text: str,
    ppem: int,
    variations: dict[str, float] | None = None,
    *,
    target_height: int,
    baseline_y: int,
) -> Image.Image:
    typeface = skia.Typeface.MakeFromFile(str(font_path))
    if typeface is None:
        raise RuntimeError(f"Skia could not load font: {font_path}")
    if variations:
        typeface = _apply_variations(typeface, variations)

    font = skia.Font(typeface, float(ppem))
    font.setSubpixel(True)

    blob = skia.TextBlob.MakeFromString(text, font)
    bounds = blob.bounds()
    width = max(int(bounds.width()) + PADDING * 2, 1)

    surface = skia.Surface(width, target_height)
    with surface as canvas:
        canvas.clear(skia.ColorWHITE)
        paint = skia.Paint()
        paint.setColor(skia.ColorBLACK)
        paint.setAntiAlias(True)
        canvas.drawTextBlob(blob, PADDING, baseline_y, paint)

    info = skia.ImageInfo.Make(
        width, target_height, skia.kRGBA_8888_ColorType, skia.kUnpremul_AlphaType
    )
    buffer = bytearray(width * target_height * 4)
    if not surface.readPixels(info, buffer, width * 4, 0, 0):
        raise RuntimeError("Skia surface.readPixels failed")
    return Image.frombytes("RGBA", (width, target_height), bytes(buffer)).convert("RGB")


def _apply_variations(typeface, variations: dict[str, float]):
    Coord = skia.FontArguments.VariationPosition.Coordinate
    coord_list = skia.FontArguments.VariationPosition.Coordinates(
        [Coord(_tag_to_int(tag), float(val)) for tag, val in variations.items()]
    )
    pos = skia.FontArguments.VariationPosition(coord_list)
    args = skia.FontArguments()
    args.setVariationDesignPosition(pos)
    return typeface.makeClone(args)


def _tag_to_int(tag: str) -> int:
    if len(tag) != 4:
        raise ValueError(f"axis tag must be 4 chars: {tag!r}")
    b = tag.encode("ascii")
    return (b[0] << 24) | (b[1] << 16) | (b[2] << 8) | b[3]



#!/usr/bin/env python3
#
# Copyright 2016 The Fontbakery Authors
# Copyright 2017 The Google Font Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import print_function

import argparse
import os
import sys
import requests

if int(sys.version[0]) == 2:
    import urlparse
elif int(sys.version[0]) == 3:
    import urllib.parse as urlparse
from gfmetadata import FamilyProto
from gfmetadata import text_format

description = (
    "This script compares the info on local METADATA.pb files"
    " with data fetched from the Google Fonts Developer API.\n\n"
    " In order to use it you need to provide an API key."
)
parser = argparse.ArgumentParser(description=description)
parser.add_argument("key", help="Key from Google Fonts Developer API")
parser.add_argument(
    "repo", help=("Directory tree that contains" " directories with METADATA.pb files.")
)
parser.add_argument(
    "--cache",
    help=("Directory to store a copy" " of the files in the fonts developer API."),
    default="/tmp/gftools-compare-git-api",
)
parser.add_argument(
    "--verbose", help="Print additional information", action="store_true"
)
parser.add_argument("--ignore-copy-existing-ttf", action="store_true")
parser.add_argument(
    "--autofix", help="Apply automatic fixes to files.", action="store_true"
)
parser.add_argument(
    "--api", help="Domain string to use to request.", default="fonts.googleapis.com"
)


def get_cache_font_path(cache_dir, fonturl):
    urlparts = urlparse.urlparse(fonturl)
    cache_dir = os.path.join(
        cache_dir, urlparts.netloc, os.path.dirname(urlparts.path).strip("/")
    )
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    fontname = os.path.basename(fonturl)
    return os.path.join(cache_dir, fontname)


def getVariantName(item):
    if item.style == "normal" and item.weight == 400:
        return "regular"

    name = ""
    if item.weight != 400:
        name = str(item.weight)

    if item.style != "normal":
        name += item.style

    return name


API_URL = "https://www.googleapis.com/webfonts/v1/webfonts?key={}"


def main(args=None):
    args = parser.parse_args(args)
    response = requests.get(API_URL.format(args.key))
    try:
        webfontList = response.json()["items"]
        webfontListFamilyNames = [item["family"] for item in webfontList]
    except (ValueError, KeyError):
        sys.exit(
            "Unable to load and parse" " list of families from Google Web Fonts API."
        )

    for dirpath, dirnames, filenames in os.walk(args.repo):
        metadata_path = os.path.join(dirpath, "METADATA.pb")
        if not os.path.exists(metadata_path):
            continue

        metadata = FamilyProto()
        text_data = open(metadata_path, "rb").read()
        text_format.Merge(text_data, metadata)
        try:
            family = metadata.name
        except KeyError:
            print(
                ('ERROR: "{}" does not contain' " familyname info.").format(
                    metadata_path
                ),
                file=sys.stderr,
            )
            continue

        try:
            index = webfontListFamilyNames.index(family)
            webfontsItem = webfontList[index]
        except ValueError:
            print(
                (
                    'ERROR: Family "{}" could not be found' " in Google Web Fonts API."
                ).format(family)
            )
            continue

        webfontVariants = []
        log_messages = []
        for variant, fonturl in webfontsItem["files"].items():
            cache_font_path = get_cache_font_path(args.cache, fonturl)
            webfontVariants.append(variant)

            if args.ignore_copy_existing_ttf and os.path.exists(cache_font_path):
                continue

            with open(cache_font_path, "w") as fp:
                found = False
                for font in metadata.fonts:
                    if getVariantName(font) == variant:
                        found = True
                        if args.verbose:
                            print(
                                'Downloading "{}"'
                                ' as "{}"'.format(fonturl, font.filename)
                            )

                        # Saving:
                        fp.write(requests.get(fonturl).text)

                        # Symlinking:
                        src = cache_font_path
                        dst_dir = os.path.dirname(cache_font_path)
                        dst = os.path.join(dst_dir, font.filename)
                        if not os.path.exists(dst):
                            os.symlink(src, dst)
                if not found:
                    print(
                        (
                            "ERROR: Google Fonts API references"
                            " a '{}' variant which is not declared"
                            " on local '{}'."
                        ).format(variant, metadata_path)
                    )

        for subset in webfontsItem["subsets"]:
            if subset == "menu":
                # note about Google Web Fonts:
                # Menu subsets are no longer generated offline.
                continue

            if subset not in metadata.subsets:
                print(
                    ('ERROR: "{}" ' 'lacks subset "{}" in git.').format(family, subset),
                    file=sys.stderr,
                )
            else:
                if args.verbose:
                    print(('OK: "{}" ' 'subset "{}" in sync.').format(family, subset))

        for subset in metadata.subsets:
            if subset != "menu" and subset not in webfontsItem["subsets"]:
                print(
                    ('ERROR: "{}" ' 'lacks subset "{}" in API.').format(family, subset),
                    file=sys.stderr,
                )

        if metadata.category == "SANS_SERIF":  # That's fine :-)
            category = "sans-serif"
        else:
            category = metadata.category.lower()

        if category != webfontsItem["category"]:
            print(
                (
                    'ERROR: "{}" category "{}" in git'
                    ' does not match category "{}"'
                    " in API."
                ).format(family, metadata.category, webfontsItem["category"])
            )
        else:
            if args.verbose:
                print(
                    ('OK: "{}" ' 'category "{}" in sync.').format(
                        family, metadata.category
                    )
                )

        for variant in webfontVariants:
            try:
                idx = [getVariantName(f) for f in metadata.fonts].index(variant)
                repoFileName = metadata.fonts[idx].filename

                fonturl = webfontsItem["files"][variant]
                fontpath = get_cache_font_path(args.cache, fonturl)

                import hashlib

                google_md5 = hashlib.md5(open(fontpath, "rb").read()).hexdigest()
                data = open(os.path.join(dirpath, repoFileName), "rb").read()
                repo_md5 = hashlib.md5(data).hexdigest()

                if repo_md5 == google_md5:
                    log_messages.append(
                        [variant, "OK", '"{}" in sync'.format(repoFileName)]
                    )
                else:
                    log_messages.append(
                        [
                            variant,
                            "ERROR",
                            (
                                '"{}" checksum mismatch, file'
                                " in API does not match file"
                                " in git."
                            ).format(repoFileName),
                        ]
                    )

            except ValueError:
                log_messages.append(
                    [
                        variant,
                        "ERROR",
                        ('"{}" available in API but' " not in git.").format(
                            font.filename
                        ),
                    ]
                )

        for font in metadata.fonts:
            variant = getVariantName(font)
            try:
                webfontVariants.index(variant)
            except ValueError:
                log_messages.append(
                    [
                        variant,
                        "ERROR",
                        ('"{}" available in git but' " not in API.").format(
                            font.filename
                        ),
                    ]
                )

        # Sort all the messages by their respective
        # metadataFileName and print them:
        for message in sorted(log_messages, key=lambda x: x[0].lower()):
            variant, status, text = message
            if status == "OK":
                if args.verbose:
                    print("{}: {}".format(status, text))
            else:
                print("{}: {}".format(status, text), file=sys.stderr)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tool to print subsets supported by a given font file.

"""
from __future__ import print_function
import argparse
import os

from gfsubsets import SubsetsInFont

parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "--min_pct",
    type=int,
    default=0,
    help="What percentage of subset codepoints have to be supported"
    " for a non-ext subset.",
)
parser.add_argument(
    "--min_pct_ext",
    type=int,
    default=0,
    help="What percentage of subset codepoints have to be supported"
    " for an -ext subset.",
)
parser.add_argument("fonts", nargs="+", metavar="FONT")


def main(args=None):
    args = parser.parse_args(args)
    for arg in args.fonts:
        subsets = SubsetsInFont(arg, args.min_pct, args.min_pct_ext)
        for subset, available, total in subsets:
            print("%s %s %d/%d" % (os.path.basename(arg), subset, available, total))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2016 The Font Bakery Authors.
# Copyright 2017 The Google Font Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import argparse
import glob
import logging
from fontTools import ttLib


# set up some command line argument processing
parser = argparse.ArgumentParser(description="Print vendorID" " of TTF files")
parser.add_argument(
    "arg_filepaths",
    nargs="+",
    help="font file path(s) to check." " Wildcards like *.ttf are allowed.",
)


def main(args=None):
    # set up a basic logging config
    # to include timestamps
    # log_format = '%(asctime)s %(levelname)-8s %(message)s'
    global font
    log_format = "%(levelname)-8s %(message)s  "
    logger = logging.getLogger()
    handler = logging.StreamHandler()
    formatter = logging.Formatter(log_format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    args = parser.parse_args(args)

    # ------------------------------------------------------
    logging.debug("Checking each file is a ttf")
    fonts_to_check = []
    for arg_filepath in sorted(args.arg_filepaths):
        # use glob.glob to accept *.ttf
        for fullpath in glob.glob(arg_filepath):
            file_path, file_name = os.path.split(fullpath)
            if file_name.endswith(".ttf"):
                fonts_to_check.append(fullpath)
            else:
                logging.warning(
                    "Skipping '{}' as it does not seem "
                    "to be valid TrueType font file.".format(file_name)
                )
    fonts_to_check.sort()

    if fonts_to_check == []:
        logging.error("None of the fonts are valid TrueType files!")

    # ------------------------------------------------------
    for font_file in fonts_to_check:
        font = ttLib.TTFont(font_file)
        logging.info("OK: {} opened with fontTools".format(font_file))

        # ----------------------------------------------------
        vid = font["OS/2"].achVendID
        print("[{}]: {}".format(vid, font_file))


__author__ = "The Google Fonts Tools Authors"
if __name__ == "__main__":
    main()



from pathlib import Path
from typing import List
from gftools.builder.file import File
from gftools.builder.operations import FontmakeOperationBase
import glyphsLib
import os
import gftools.builder
from functools import cached_property
from glyphsLib.builder import UFOBuilder
from ninja.ninja_syntax import Writer, escape_path, escape
from fontTools.designspaceLib import InstanceDescriptor


class InstantiateUFO(FontmakeOperationBase):
    description = "Create instance UFOs from a Glyphs or designspace file"
    rule = 'fontmake -i "$instance_name" -o ufo $fontmake_type $in $args'

    def validate(self):
        # Ensure there is an instance name
        if "instance_name" not in self.original:
            raise ValueError("No instance name specified")
        # Ensure the instance is defined in the font
        desired = self.original["instance_name"]
        if "target" not in self.original and not self.relevant_instance:
            raise ValueError(
                f"Instance {desired} not found in {self.first_source.path}"
            )

    @cached_property
    def relevant_instance(self):
        desired = self.original["instance_name"]
        relevant_instance = [
            i
            for i in self.first_source.instances
            if i.name == desired or i.familyName + " " + i.styleName == desired
        ]
        if len(relevant_instance) == 0:
            return None
        return relevant_instance[0]

    @property
    def instance_dir(self):
        return Path(self.first_source.path).parent / "instance_ufos"

    @property
    def targets(self):
        if "target" in self.original:
            return [File(self.original["target"])]
        instance = self.relevant_instance
        assert instance is not None
        assert instance.filename is not None
        # if self.first_source.is_glyphs:
        return [
            File(
                str(self.instance_dir / (os.path.basename(instance.filename) + ".json"))
            )
        ]
        # return [ File(instance.filename) ]

    @property
    def variables(self):
        vars = super().variables
        vars["args"] += " --ufo-structure=json "
        if self.first_source.is_glyphs:
            vars["args"] += f"--instance-dir {escape_path(str(self.instance_dir))}"
        else:
            vars["args"] += f"--output-dir {escape_path(str(self.instance_dir))}"
        vars["instance_name"] = escape(self.original["instance_name"])
        if self.original.get("glyphData") is not None:
            for glyphData in self.original["glyphData"]:
                vars["args"] += f" --glyph-data {escape_path(glyphData)}"
        return vars

    def set_target(self, target: File):
        raise ValueError("Cannot set target on InstantiateUFO")



"""functionality for running fontc via gftools

gftools has a few special flags that allow it to use fontc, an alternative
font compiler (https://github.com/googlefonts/fontc).

This module exists to keep the logic related to fontc in one place, and not
dirty up everything else.
"""

from argparse import Namespace
from pathlib import Path
from typing import Union
import time

from gftools.builder.file import File
from gftools.builder.operations.fontc import set_global_fontc_path


class FontcArgs:
    # init with 'None' returns a default obj where everything is None
    def __init__(self, args: Union[Namespace, None]) -> None:
        if not args:
            self.simple_output_path = None
            self.fontc_bin_path = None
            self.single_source = None
            return
        self.simple_output_path = abspath(args.experimental_simple_output)
        self.fontc_bin_path = abspath(args.experimental_fontc)
        self.single_source = args.experimental_single_source
        if self.fontc_bin_path:
            if not self.fontc_bin_path.is_file():
                raise ValueError(f"fontc does not exist at {self.fontc_bin_path}")
            set_global_fontc_path(self.fontc_bin_path)

    @property
    def use_fontc(self) -> bool:
        return self.fontc_bin_path is not None

    def build_file_name(self) -> str:
        if self.fontc_bin_path or self.simple_output_path:
            # if we're running for fontc we want uniquely named build files,
            # to ensure they don't collide
            return f"build-{time.time_ns()}.ninja"
        else:
            # otherwise just ues the default name
            return "build.ninja"

    # update the config dictionary based on our special needs
    def modify_config(self, config: dict):
        if self.single_source:
            filtered_sources = [s for s in config["sources"] if self.single_source in s]
            n_sources = len(filtered_sources)
            if n_sources != 1:
                raise ValueError(
                    f"--exerimental-single-source {self.single_source} must match exactly one of {config['sources']} (matched {n_sources}) "
                )
            config["sources"] = filtered_sources

        if self.fontc_bin_path or self.simple_output_path:
            # we stash this flag here to pass it down to the recipe provider
            config["use_fontc"] = self.fontc_bin_path
            config["buildWebfont"] = False
            config["buildSmallCap"] = False
            config["splitItalic"] = False
            config["cleanUp"] = True
            # disable running ttfautohint, because we had a segfault
            config["autohintTTF"] = False
            extra_args = config.get("extraFontmakeArgs") or ""
            extra_args += " --drop-implied-oncurves"
            config["extraFontmakeArgs"] = extra_args
            # override config to turn not build instances if we're variable
            if self.will_build_variable_font(config):
                config["buildStatic"] = False
            # if the font doesn't explicitly request CFF, just build TT outlines
            # if the font _only_ wants CFF outlines, we will try to build them
            # ( but fail on fontc for now) (but is this even a thing?)
            elif config.get("buildTTF", True):
                config["buildOTF"] = False
        if self.simple_output_path:
            output_dir = str(self.simple_output_path)
            # we dump everything into one dir in this case
            config["outputDir"] = str(output_dir)
            config["ttDir"] = str(output_dir)
            config["otDir"] = str(output_dir)
            config["vfDir"] = str(output_dir)

    def will_build_variable_font(self, config: dict) -> bool:
        # if config explicitly says dont build variable, believe it
        if not config.get("buildVariable", True):
            return False

        source = File(config["sources"][0])
        return source.is_variable


def abspath(path: Union[Path, None]) -> Union[Path, None]:
    return path.resolve() if path else None



#!/usr/bin/env python3
#
# Copyright 2017 The Google Font Tools Authors
# Copyright 2010, Google Inc.
# Author: Dave Crossland (dave@understandinglimited.com)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""
A Python script for printing nametables to stdout.

e.g:

Check nametables of fonts in collection:
gftools check-name [fonts]

Output in csv format
gftools check-name [fonts] --csv
"""
from argparse import ArgumentParser, RawTextHelpFormatter
import csv
import sys
from fontTools.ttLib import TTFont
import tabulate
import ntpath

parser = ArgumentParser(description=__doc__, formatter_class=RawTextHelpFormatter)
parser.add_argument("fonts", nargs="+", help="Fonts in OpenType (TTF/OTF) format")
parser.add_argument("--csv", default=False, action="store_true")


def printInfo(rows, save=False):
    header = [r[0] for r in rows[0]]
    t = []
    for row in rows:
        t.append([r[1] for r in row])

    if save:
        writer = csv.writer(sys.stdout)
        writer.writerows([header])
        writer.writerows(t)
        sys.exit(0)
    else:
        print(tabulate.tabulate(t, header, tablefmt="plain"))


def main(args=None):
    args = parser.parse_args(args)

    rows = []
    for font_filename in args.fonts:
        font = TTFont(font_filename)
        for field in font["name"].names:
            enc = field.getEncoding()
            rows.append(
                [
                    ("Font", ntpath.basename(font_filename)),
                    ("platformID", field.platformID),
                    ("encodingID", field.platEncID),
                    ("languageID", field.langID),
                    ("nameID", field.nameID),
                    ("nameString", field.toUnicode()),
                ]
            )

    if args.csv:
        printInfo(rows, save=True)
    else:
        printInfo(rows)


if __name__ == "__main__":
    main()



from ttfautohint import ttfautohint
from fontTools.ttLib import TTFont
from ttfautohint.options import parse_args as ttfautohint_parse_args
from gftools.utils import primary_script
import sys

AUTOHINT_SCRIPTS = [
    "adlm",
    "arab",
    "armn",
    "avst",
    "bamu",
    "beng",
    "buhd",
    "cakm",
    "cans",
    "cari",
    "cher",
    "copt",
    "cprt",
    "cyrl",
    "deva",
    "dsrt",
    "ethi",
    "geor",
    "geok",
    "glag",
    "goth",
    "grek",
    "gujr",
    "guru",
    "hebr",
    "hmnp",
    "kali",
    "khmr",
    "khms",
    "knda",
    "lao",
    "latn",
    "latb",
    "latp",
    "lisu",
    "mlym",
    "medf",
    "mong",
    "mymr",
    "nkoo",
    "olck",
    "orkh",
    "osge",
    "osma",
    "rohg",
    "saur",
    "shaw",
    "sinh",
    "sund",
    "taml",
    "tavt",
    "telu",
    "tfng",
    "thai",
    "vaii",
    "yezi",
]


def autohint_script_tag(ttFont, discount_latin=False):
    script = primary_script(ttFont, ignore_latin=discount_latin)
    if script in AUTOHINT_SCRIPTS:
        return script
    return


def autohint(infile, outfile, args=None, add_script=False, discount_latin=False):
    font = TTFont(infile)
    if not args:
        args = []
        if isinstance(add_script, str) and add_script != "auto":
            args.append("-D" + add_script)
        elif add_script:  # True or "auto"
            script = autohint_script_tag(font, discount_latin=discount_latin)
            if script:
                args.append("-D" + script)
    args_dict = ttfautohint_parse_args([infile, outfile, *args])
    if not args_dict:
        raise ValueError("Could not parse arguments")
    ttfautohint(**args_dict)



from pathlib import Path
from typing import List
from gftools.builder.operations import OperationBase

_FONTC_PATH = None


# should only be called once, from main, before doing anything else. This is a
# relatively non-invasive way to smuggle this value into FontcOperationBase
def set_global_fontc_path(path: Path):
    global _FONTC_PATH
    assert _FONTC_PATH is None, "set_global_fontc_path should only be called once"
    _FONTC_PATH = path


class FontcOperationBase(OperationBase):
    @property
    def variables(self):
        vars = super().variables
        vars["fontc_path"] = _FONTC_PATH
        args = vars.get("args")
        if args:
            vars["args"] = rewrite_fontmake_args_for_fontc(args)

        return vars


def rewrite_fontmake_args_for_fontc(args: str) -> str:
    out_args = []
    arg_list = args.split()
    # reverse so we can pop in order
    arg_list.reverse()
    while arg_list:
        out_args.append(rewrite_one_arg(arg_list))
    return " ".join(out_args)


# remove next arg from the front of the list and return its fontc equivalent
def rewrite_one_arg(args: List[str]) -> str:
    next_ = args.pop()
    if next_ == "--filter":
        filter_ = args.pop()
        # this means 'retain filters defined in UFO', which... do we even support
        # that in fontc?
        if filter_ == "...":
            return ""
        elif filter_ == "FlattenComponentsFilter":
            return "--flatten-components"
        elif filter_ == "DecomposeTransformedComponentsFilter":
            return "--decompose-transformed-components"
        elif "DecomposeComponentsFilter" in filter_:
            # this is sometimes spelled with the full qualified path name as
            # --filter "ufo2ft.filters.decomposeComponents::DecomposeComponentsFilter"
            # e.g. in Jaquard12.glyphs so we use `in` instead of `filter_ == ...`
            return "--decompose-components"
        else:
            # glue the filter back together for better reporting below
            next_ = f"{next_} {filter_}"
    elif next_ == "--no-production-names":
        return next_
    elif next_ == "--verbose":
        log_level = python_to_rust_log_level(args.pop().strip())
        return f"--log={log_level}"
    elif next_ == "--drop-implied-oncurves" or next_ == "--keep-overlaps":
        # this is our default behaviour so no worries
        return ""
    elif next_ == "--no-check-compatibility":
        # we don't have an equivalent
        return ""
    raise ValueError(f"unknown fontmake arg '{next_}'")


def python_to_rust_log_level(py_level: str):
    if py_level == "WARNING":
        return "WARN"
    elif py_level == "CRITICAL":
        return "ERROR"
    else:
        return py_level



from gftools.builder.operations import FontmakeOperationBase


class BuildOTF(FontmakeOperationBase):
    description = "Build a OTF from a source file"
    rule = "fontmake --output-path $out -o otf $fontmake_type $in $args"

    def validate(self):
        if not self.first_source.exists():
            # We can't check this file (assume it was generated as part of the
            # build), so user is on their own.
            return
        if self.first_source.is_glyphs and len(self.first_source.gsfont.masters) > 1:
            raise ValueError(
                f"Cannot build a static font from {self.first_source.path}"
            )
        if (
            self.first_source.designspace
            and len(self.first_source.designspace.sources) > 1
        ):
            raise ValueError(
                f"Cannot build a static font from {self.first_source.path}"
            )



#!/usr/bin/env python3
"""Create or author Google Fonts axisregistry {AXIS_NAME}.textproto files."""

import sys
import argparse
from gfmetadata import AxisProto, FallbackProto
from gfmetadata import text_format
from fontTools.ttLib import TTFont

parser = argparse.ArgumentParser(
    prog="gftools add-axis",
    description=__doc__,
)

parser.add_argument("font", type=str, help="The font file to the axis values from.")


class ProgramAbortError(Exception):
    pass


class UserAbortError(Exception):
    pass


def _get_fvar_axis(name_table, fvar_table):
    axes = []
    for axis in fvar_table.axes:
        axes.append(
            (
                axis,
                f"{name_table.getName(axis.axisNameID, 3, 1, 0x0409)} {axis.axisTag}",
            )
        )
    axes.sort(key=lambda a: a[0].axisTag)
    choices = "\n".join(
        [f"  {index}: {label}" for index, (_, label) in enumerate(axes)]
    )
    question = "Found axes:\n" f"{choices}" "\n" "pick one by number (e.g. 0), q=quit:"
    while True:
        try:
            answer = input(question).strip()
            if answer == "q":
                raise UserAbortError()
            index = int(answer)  # raises ValueError
            fvar_axis, _ = axes[index]  # raises IndexError
        except (ValueError, IndexError):
            # must try again
            continue
        print(f"You picked: {fvar_axis.axisTag}.")
        return fvar_axis


def _get_fallbacks_gen(name_table, stat_axis_index, AxisValue):
    for stat_axis_value in AxisValue:
        if stat_axis_value.Format in (1, 3):
            if stat_axis_value.AxisIndex == stat_axis_index:
                yield (
                    name_table.getName(stat_axis_value.ValueNameID, 3, 1, 0x0409),
                    stat_axis_value.Value,
                )
        elif stat_axis_value.Format == 4:
            for avr in stat_axis_value.AxisValueRecord:
                if avr.AxisIndex == stat_axis_index:
                    yield (
                        name_table.getName(stat_axis_value.ValueNameID, 3, 1, 0x0409),
                        avr.Value,
                    )
        else:
            print(
                f"SKIP STAT AxisValue can't handel Format {stat_axis_value.Format} "
                f"({name_table.getName(stat_axis_value.ValueNameID, 3, 1, 0x0409)})"
            )


def add_axis(font: str):
    axis_proto = AxisProto()
    ttFont = TTFont(font)
    name_table = ttFont["name"]
    try:
        fvar_table = ttFont["fvar"]
    except KeyError:
        raise ProgramAbortError("No fvar present")
    fvar_axis = _get_fvar_axis(name_table, fvar_table)

    # Axis tag
    axis_proto.tag = fvar_axis.axisTag
    # Display name for axis, e.g. "Optical size" for 'opsz'
    # Like 'Name' in
    # https://docs.microsoft.com/en-us/typography/opentype/spec/dvaraxistag_opsz
    # name_table.getName(
    #           NameID,
    #           <Platform ID: 3 = Windows>,
    #           <encodingID, Platform-specific encoding ID: 1 = Unicode BMP>,
    #           <Language ID: 0x0409 = 1033 = en_us>)
    axis_proto.display_name = (
        f"{name_table.getName(fvar_axis.axisNameID, 3, 1, 0x0409)}"
    )
    # Lower bound for the axis
    axis_proto.min_value = fvar_axis.minValue
    # The default position to use and to prefer for exemplars
    axis_proto.default_value = fvar_axis.defaultValue
    # Upper bound for the axis
    axis_proto.max_value = fvar_axis.maxValue
    # Input values for this axis must aligned to 10^precision
    axis_proto.precision = 1  # ask user?
    # Short descriptive paragraph
    axis_proto.description = (  # ask user?
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod"
        " tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim"
        " veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea"
        " commodo consequat. Duis aute irure dolor in reprehenderit in voluptate"
        " velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint"
        " occaecat cupidatat non proident, sunt in culpa qui officia deserunt"
        " mollit anim id est laborum."
    )

    fallback_proto = FallbackProto()
    fallback_proto.name = "Default"
    fallback_proto.value = fvar_axis.defaultValue
    axis_proto.fallback.append(fallback_proto)
    # Is the axis fallback only?
    axis_proto.fallback_only = False

    text_proto = text_format.MessageToString(
        axis_proto, as_utf8=True, use_index_order=True
    )
    filename = f"{axis_proto.display_name.lower()}.textproto"
    with open(filename, "x") as f:
        f.write(text_proto)
    print(f"DONE create {filename}!")


def main(args=None):
    try:
        args = parser.parse_args(args)
        add_axis(args.font)
    except UserAbortError:
        print("Aborted by user!")
        sys.exit(1)
    except ProgramAbortError as e:
        print(f"Aborted by program: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2017 The Fontbakery Authors
# Copyright 2017 The Google Font Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import argparse
import os
import tabulate
from fontTools import ttLib
from gftools.constants import NAMEID_COPYRIGHT_NOTICE, PLATID_STR

parser = argparse.ArgumentParser(description="Print out copyright" " nameIDs strings")
parser.add_argument("font", nargs="+")
parser.add_argument(
    "--csv",
    default=False,
    action="store_true",
    help="Output data in comma-separate-values" " (CSV) file format",
)


def main(args=None):
    args = parser.parse_args(args)

    rows = []
    for font in args.font:
        ttfont = ttLib.TTFont(font)
        for name in ttfont["name"].names:
            if name.nameID != NAMEID_COPYRIGHT_NOTICE:
                continue

            value = name.string.decode(name.getEncoding()) or ""
            rows.append(
                [
                    os.path.basename(font),
                    value,
                    len(value),
                    "{} ({})".format(
                        name.platformID, PLATID_STR.get(name.platformID, "?")
                    ),
                ]
            )

    header = ["filename", "copyright notice", "char length", "platformID"]

    def as_csv(rows):
        import csv
        import sys

        writer = csv.writer(
            sys.stdout, delimiter="|", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        writer.writerows([header])
        writer.writerows(rows)
        sys.exit(0)

    if args.csv:
        as_csv(rows)

    print("")  # some spacing
    print(tabulate.tabulate(rows, header, tablefmt="pipe"))
    print("")  # some spacing


if __name__ == "__main__":
    """Example usage:

    gftools check-copyright-notices ~/fonts/*/*/*ttf --csv > ~/notices.txt;
    """
    main()



#!/usr/bin/env python3
# Copyright 2016 The Font Bakery Authors.
# Copyright 2017 The Google Font Tools Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Update specific nameIDs in a collection of fonts with new strings.

Examples:

$ gftools update-nameids -c="Copyright 2016" font.ttf
$ gftools update-nameids -v="4.000" -ul="http://license.org" [fonts.ttf]

if you need to change the name or style of a collection of font families,
use gftools nametable-from-filename instead.
"""
from __future__ import print_function
from fontTools.ttLib import TTFont
from argparse import ArgumentParser, RawTextHelpFormatter

NAME_IDS = {
    0: "copyright",
    3: "uniqueid",
    5: "version",
    7: "trademark",
    8: "manufacturer",
    9: "designer",
    10: "description",
    11: "urlvendor",
    12: "urldesigner",
    13: "license",
    14: "urllicense",
}


def swap_name(field, font_name_field, new_name):
    """Replace a font's name field with a new name"""
    enc = font_name_field.getName(*field).getEncoding()
    text = font_name_field.getName(*field).toUnicode()
    text = new_name
    font_name_field.setName(text, *field)


def update_field(arg, args, fields, nametable):
    if hasattr(args, arg):
        text = getattr(args, arg)
        if text:
            swap_name(fields, nametable, text)


parser = ArgumentParser(description=__doc__, formatter_class=RawTextHelpFormatter)
parser.add_argument("fonts", nargs="+")
parser.add_argument("-c", "--copyright", type=str, help="Update copyright string")
parser.add_argument("-u", "--uniqueid", type=str, help="Update uniqueid string")
parser.add_argument("-v", "--version", type=str, help="Update version string")
parser.add_argument("-t", "--trademark", type=str, help="Update trademark string")
parser.add_argument("-m", "--manufacturer", type=str, help="Update manufacturer string")
parser.add_argument("-d", "--designer", type=str, help="Update designer string")
parser.add_argument(
    "-desc", "--description", type=str, help="Update description string"
)
parser.add_argument("-l", "--license", type=str, help="Update license string")
parser.add_argument("-uv", "--urlvendor", type=str, help="Update url vendor string")
parser.add_argument("-ud", "--urldesigner", type=str, help="Update url vendor string")
parser.add_argument("-ul", "--urllicense", type=str, help="Update url license string")


def main(args=None):
    args = parser.parse_args(args)

    for font_path in args.fonts:
        font = TTFont(font_path)
        for field in font["name"].names:
            fields = (field.nameID, field.platformID, field.platEncID, field.langID)
            if field.nameID in NAME_IDS:
                update_field(NAME_IDS[field.nameID], args, fields, font["name"])

        font.save(font_path + ".fix")
        print("font saved %s.fix" % font_path)


if __name__ == "__main__":
    main()



import contextlib
import os
import re
import selectors
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Dict, List
from venv import EnvBuilder

import git
import yaml
from rich import progress
from rich.progress import Progress

from gfmetadata import FamilyProto

# Python <3.11
if not hasattr(contextlib, "chdir"):
    from contextlib import AbstractContextManager

    class chdir(AbstractContextManager):
        """Non thread-safe context manager to change the current working directory."""

        def __init__(self, path):
            self.path = path
            self._old_cwd = []

        def __enter__(self):
            self._old_cwd.append(os.getcwd())
            os.chdir(self.path)

        def __exit__(self, *excinfo):
            os.chdir(self._old_cwd.pop())

    contextlib.chdir = chdir


class GitRemoteProgress(git.RemoteProgress):
    OP_CODES = [
        "BEGIN",
        "CHECKING_OUT",
        "COMPRESSING",
        "COUNTING",
        "END",
        "FINDING_SOURCES",
        "RECEIVING",
        "RESOLVING",
        "WRITING",
    ]
    OP_CODE_MAP = {
        getattr(git.RemoteProgress, _op_code): _op_code for _op_code in OP_CODES
    }

    def __init__(self, progressbar, task, name) -> None:
        super().__init__()
        self.progressbar = progressbar
        self.task = task
        self.name = name
        self.curr_op = None

    @classmethod
    def get_curr_op(cls, op_code: int) -> str:
        """Get OP name from OP code."""
        # Remove BEGIN- and END-flag and get op name
        op_code_masked = op_code & cls.OP_MASK
        return cls.OP_CODE_MAP.get(op_code_masked, "?").title()

    def update(
        self,
        op_code: int,
        cur_count: str | float,
        max_count: str | float | None = None,
        message: str | None = "",
    ) -> None:
        if not self.progressbar:
            return
        # Start new bar on each BEGIN-flag
        if op_code & self.BEGIN:
            self.curr_op = self.get_curr_op(op_code)
            # logger.info("Next: %s", self.curr_op)
            self.progressbar.update(
                self.task,
                description="[yellow] " + self.curr_op + " " + self.name,
                total=max_count,
            )

        self.progressbar.update(
            task_id=self.task,
            completed=cur_count,
            message=message,
        )


def find_config_yaml(source_dir: Path):
    configs = []
    for path in source_dir.glob("sources/*.y*l"):
        if not (str(path).endswith(".yaml") or str(path).endswith(".yml")):
            continue
        content = yaml.load(path.read_text(), Loader=yaml.Loader)
        if "sources" not in content:
            continue
        configs.append(path)
    if configs:
        return configs[0]


def find_sources(source_dir: Path) -> List[Path]:
    # Extensions in order of preference
    for extension in [".glyphs", ".glyphspackage", ".designspace", ".ufo"]:
        sources = list(source_dir.glob("sources/*" + extension))
        if sources:
            return sources
    return []


class SourceBuilder:
    def __init__(
        self,
        destination: Path,
        family_path: Path,
        metadata: FamilyProto,
        their_venv: bool = False,
    ):
        self.destination = destination
        self.family_path = family_path
        self.metadata = metadata
        self.name = metadata.name
        self.their_venv = their_venv
        self.progressbar = None
        self.source_dir = tempfile.TemporaryDirectory()

    def setup_venv(self, source_dir: Path):
        venv_task = self.progressbar.add_task(
            "[yellow]Setup venv", total=500, visible=False
        )
        self.progressbar.update(
            venv_task,
            description="[yellow] Setting up venv for " + self.name + "...",
            completed=0,
            visible=True,
        )
        if (source_dir / "Makefile").exists():
            with contextlib.chdir(source_dir):
                # self.progressbar.console.print(
                #     "[yellow]Running make venv in " + str(source_dir)
                # )
                rc = self.run_command_with_callback(
                    ["make", "venv"],
                    lambda line: self.progressbar.update(venv_task, advance=1),
                )
        elif (source_dir / "requirements.txt").exists():
            builder = EnvBuilder(system_site_packages=False, with_pip=True)
            builder.create(str(source_dir / "venv"))
            self.progressbar.update(venv_task, completed=10)
            # self.progressbar.console.print(
            #     "[yellow]Running pip install in " + str(source_dir)
            # )
            with contextlib.chdir(source_dir):
                rc = self.run_command_with_callback(
                    ["venv/bin/pip", "install", "-r", "requirements.txt"],
                    lambda line: self.progressbar.update(venv_task, advance=1),
                )
        else:
            raise ValueError(
                "--their-venv was provided but no Makefile or requirements.txt upstream"
            )
        if rc != 0:
            self.progressbar.console.print(
                "[red]Error setting up venv for " + self.name
            )
            raise ValueError("Venv setup failed")
        self.progressbar.remove_task(venv_task)

    def local_overrides(
        self, upstream: Path, downstream: Path, overrides: Dict[str, str]
    ):
        for source, dest in overrides.items():
            if (downstream / source).exists():
                dest_path = upstream / dest
                os.makedirs(dest_path.parent, exist_ok=True)
                self.progressbar.console.print("[grey]Using our " + source)
                shutil.copy(downstream / source, dest_path)

    def build(self):
        with Progress(
            progress.TimeElapsedColumn(),
            progress.TextColumn("[progress.description]{task.description}"),
            progress.BarColumn(),
            progress.TextColumn("{task.completed}/{task.total}"),
            progress.TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            progress.TimeRemainingColumn(),
        ) as self.progressbar:
            with tempfile.TemporaryDirectory() as source_dir:
                source_dir = Path(source_dir)
                self.clone_source(source_dir)
                self.local_overrides(
                    source_dir,
                    self.family_path,
                    {
                        "config.yaml": "sources/config.yaml",
                        "requirements.txt": "requirements.txt",
                    },
                )

                if not (source_dir / "sources").exists():
                    raise ValueError(f"Could not find sources directory in {self.name}")
                if self.their_venv:
                    self.setup_venv(source_dir)

                # Locate the config.yaml file or first source
                arg = find_config_yaml(source_dir)
                if not arg:
                    sources = find_sources(source_dir)
                    if not sources:
                        raise ValueError(
                            f"Could not find any sources in {self.metadata.source}"
                        )
                    arg = sources[0]

                with contextlib.chdir(source_dir):
                    if self.their_venv:
                        buildcmd = ["venv/bin/gftools-builder", str(arg)]
                    else:
                        buildcmd = ["gftools-builder", str(arg)]
                    self.run_build_command(buildcmd)
                    self.copy_files()

    def run_command_with_callback(self, cmd, callback):
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        sel = selectors.DefaultSelector()
        sel.register(process.stdout, selectors.EVENT_READ)
        sel.register(process.stderr, selectors.EVENT_READ)
        ok = True
        stdoutlines = []
        stderrlines = []
        while ok:
            for key, _val1 in sel.select():
                line = key.fileobj.readline()
                if not line:
                    ok = False
                    break
                if key.fileobj is process.stdout:
                    callback(line)
                elif key.fileobj is process.stderr:
                    stderrlines.append(line)
                else:
                    stdoutlines.append(line)
        rc = process.wait()
        if rc != 0:
            for line in stdoutlines:
                self.progressbar.console.print(line.decode("utf-8"), end="")
            for line in stderrlines:
                self.progressbar.console.print("[red]" + line.decode("utf8"), end="")
        return rc

    def run_build_command(self, buildcmd):
        build_task = self.progressbar.add_task("[green]Build " + self.name, total=1)

        def progress_callback(line):
            if m := re.match(r"^\[(\d+)/(\d+)\]", line.decode("utf8")):
                self.progressbar.update(
                    build_task, completed=int(m.group(1)), total=int(m.group(2))
                )

        if self.run_command_with_callback(buildcmd, progress_callback) != 0:
            self.progressbar.console.print("[red]Error building " + self.name)
            raise ValueError("Build failed")
        else:
            self.progressbar.console.print("[green]Built " + self.name)

    def clone_source(
        self,
        builddir: Path,
    ):
        clone_task = self.progressbar.add_task(
            "[yellow]Clone", total=100, visible=False
        )
        self.progressbar.update(
            clone_task,
            description="[yellow] Cloning " + self.name + "...",
            completed=0,
            visible=True,
        )
        git.Repo.clone_from(
            url=self.metadata.source.repository_url,
            to_path=builddir,
            depth=1,
            progress=GitRemoteProgress(self.progressbar, clone_task, self.name),
        )
        self.progressbar.remove_task(clone_task)

    def copy_files(self):
        # We are sat in the build directory
        for item in self.metadata.source.files:
            in_fp = Path(item.source_file)
            if not in_fp.exists():
                raise ValueError(
                    f"Expected to copy {item.source_file} but it was not found after build"
                )
            out_fp = Path(self.destination / item.dest_file)
            if not out_fp.parent.exists():
                os.makedirs(out_fp.parent, exist_ok=True)
            shutil.copy(in_fp, out_fp)


def build_to_directory(
    destination: Path,
    family_path: Path,
    metadata: FamilyProto,
    their_venv: bool = False,
):
    SourceBuilder(destination, family_path, metadata, their_venv).build()



#!/usr/bin/env python3
"""
gftools lang-support

Adds language support metadata to METADATA.pb files.

Usage:

# Standard usage. Does not overwrite existing data.
gftools lang-support -l ./lang/ ./ofl/noto*/METADATA.pb

# Generate a report with insights about data and potential metadata holes.
gftools lang-support -l ./lang/ -r ./ofl/noto*/METADATA.pb

"""

import argparse
from fontTools.ttLib import TTFont
from gflanguages import LoadLanguages, LoadScripts
from gfmetadata import FamilyProto
from gftools.util import google_fonts as fonts
import csv
import os

parser = argparse.ArgumentParser(
    description="Add language support metadata to METADATA.pb files"
)
parser.add_argument("--lang", "-l", help="Path to lang metadata package")
parser.add_argument(
    "--report",
    "-r",
    action="store_true",
    help="Whether to output a report of lang metadata insights",
)
parser.add_argument(
    "--sample_text_audit",
    "-s",
    action="store_true",
    help="Whether to run the sample text audit",
)
parser.add_argument("--out", "-o", help="Path to output directory for report")
parser.add_argument("metadata_files", help="Path to METADATA.pb files", nargs="+")


def _WriteCsv(path, rows):
    with open(path, "w", newline="") as csvfile:
        writer = csv.writer(
            csvfile, delimiter="\t", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        for row in rows:
            writer.writerow(row)


def _WriteReport(metadata_paths, out_dir, languages):
    rows = [
        [
            "id",
            "name",
            "lang",
            "script",
            "population",
            "ec_base",
            "ec_auxiliary",
            "ec_marks",
            "ec_numerals",
            "ec_punctuation",
            "ec_index",
            "st_fallback",
            "st_fallback_name",
            "st_masthead_full",
            "st_masthead_partial",
            "st_styles",
            "st_tester",
            "st_poster_sm",
            "st_poster_md",
            "st_poster_lg",
            "st_specimen_48",
            "st_specimen_36",
            "st_specimen_32",
            "st_specimen_21",
            "st_specimen_16",
        ]
    ]

    without_lang = []
    without_sample_text = []
    supported_without_sample_text = {}
    for metadata_path in metadata_paths:
        family = fonts.ReadProto(FamilyProto(), metadata_path)
        if len(family.languages) == 0:
            without_lang.append(family.name)
        else:
            supports_lang_with_sample_text = False
            for lang_code in family.languages:
                if languages[lang_code].HasField("sample_text"):
                    supports_lang_with_sample_text = True
                    break
            if not supports_lang_with_sample_text:
                without_sample_text.append(family.name)
        for l in family.languages:
            if (
                not languages[l].HasField("sample_text")
                and l not in supported_without_sample_text
            ):
                supported_without_sample_text[l] = languages[l]

    for lang in supported_without_sample_text.values():
        rows.append([lang.id, lang.name, lang.language, lang.script, lang.population])

    path = os.path.join(out_dir, "support.csv")
    _WriteCsv(path, rows)


def _SampleTextAudit(out_dir, languages, scripts, unused_scripts=[]):
    rows = [["id", "language", "script", "has_sample_text", "historical"]]
    # sort by script|has_sample_text|historical|id
    entries = []

    min_sample_text_languages = 0
    by_script = {}
    for l in languages.values():
        if l.script not in by_script:
            by_script[l.script] = []
        by_script[l.script].append(l)
    for script in by_script:
        if script in unused_scripts:
            continue
        languages_with_sample_text = {
            l.id
            for l in by_script[script]
            if l.HasField("sample_text")
            and not l.sample_text.HasField("fallback_language")
        }
        non_historical_languages_without_sample_text = [
            l
            for l in by_script[script]
            if not l.historical and l.id not in languages_with_sample_text
        ]
        if len(languages_with_sample_text) < 2:
            if (
                len(languages_with_sample_text) == 1
                and len(by_script[script]) > 1
                and len(non_historical_languages_without_sample_text) > 1
            ):
                min_sample_text_languages += 1
            elif len(languages_with_sample_text) == 0:
                if len(non_historical_languages_without_sample_text) > 1:
                    min_sample_text_languages += 2
                else:
                    min_sample_text_languages += 1

        if len(languages_with_sample_text) == 0 or (
            len(languages_with_sample_text) == 1
            and len([l for l in by_script[script] if not l.historical]) > 1
        ):
            for l in by_script[script]:
                entries.append(
                    {
                        "id": l.id,
                        "language": l.name,
                        "script": scripts[l.script].name,
                        "has_sample_text": l.id in languages_with_sample_text,
                        "historical": l.historical,
                    }
                )

    print(min_sample_text_languages)

    last_script = None
    entries.sort(
        key=lambda x: (
            x["script"],
            not x["has_sample_text"],
            not x["historical"],
            x["id"],
        )
    )
    for e in entries:
        if last_script is not None and e["script"] != last_script:
            rows.append([])
        rows.append(
            [
                e["id"],
                e["language"],
                e["script"],
                "X" if e["has_sample_text"] else "",
                "X" if e["historical"] else "",
            ]
        )
        last_script = e["script"]

    path = os.path.join(out_dir, "sample_text_audit.csv")
    _WriteCsv(path, rows)


def main(args=None):
    args = parser.parse_args(args)
    languages = LoadLanguages(base_dir=args.lang)
    scripts = LoadScripts(base_dir=args.lang)

    if args.report:
        assert len(argv) > 1, "No METADATA.pb files specified"
        assert args.out is not None, "No output dir specified (--out)"
        print("Writing insights report...")
        _WriteReport(argv[1:], args.out, languages)
    elif args.sample_text_audit:
        assert args.out is not None, "No output dir specified (--out)"
        print("Auditing sample text")
        seen_scripts = set()
        unused_scripts = set()
        for path in argv[1:]:
            family = fonts.ReadProto(FamilyProto(), path)
            for l in family.languages:
                seen_scripts.add(languages[l].script)
        for s in scripts:
            if s not in seen_scripts:
                unused_scripts.add(s)
        _SampleTextAudit(args.out, languages, scripts, unused_scripts)
    else:
        for path in args.metadata_files:
            family_metadata = fonts.ReadProto(FamilyProto(), path)
            if len(family_metadata.languages) > 0:
                continue
            exemplar_font_fp = os.path.join(
                os.path.dirname(path), fonts.GetExemplarFont(family_metadata).filename
            )
            exemplar_font = TTFont(exemplar_font_fp)
            supported_languages = fonts.SupportedLanguages(exemplar_font, languages)
            if family_metadata.HasField("is_noto") and family_metadata.is_noto:
                supported_languages = [
                    l for l in supported_languages if "Latn" not in l.id
                ]
            supported_languages = sorted([l.id for l in supported_languages])
            family_metadata.languages.extend(supported_languages)
            fonts.WriteMetadata(family_metadata, path)


if __name__ == "__main__":
    main()



import os
import subprocess
import tempfile
import shutil
import atexit
from collections import defaultdict
from os import chdir
from pathlib import Path
from tempfile import NamedTemporaryFile, gettempdir
from typing import Any, Dict, List, Union

from gftools.builder.fontc import FontcArgs
import networkx as nx
import strictyaml
import yaml
from fontmake.font_project import FontProject
from ninja import _program
from ninja.ninja_syntax import Writer, escape_path
from typing import Union

from gftools.builder.file import File
from gftools.builder.operations import OperationBase, OperationRegistry
from gftools.builder.operations.copy import Copy
from gftools.builder.recipeproviders import get_provider
from gftools.builder.schema import BASE_SCHEMA

Recipe = Dict[str, List[Dict[str, Any]]]


def edge_with_operation(node, operation):
    for newnode, attributes in node.items():
        if operation == attributes.get("operation"):
            return newnode
    return None


class GFBuilder:
    config: dict
    recipe: Recipe

    def __init__(
        self,
        config: Union[dict, str],
        fontc_args=FontcArgs(None),
    ):
        if isinstance(config, str):
            parentpath = Path(config).resolve().parent
            with open(config, "r") as file:
                config = file.read()
            chdir(parentpath)
            self._orig_config = config
            # StrictYAML is great for validating the input, but its strongly
            # typed nature makes it hard to work with. So we use it to
            # validate the input, but just treat it as a regular Python dict.
            try:
                strictyaml.load(config, BASE_SCHEMA)
            except Exception as e:
                raise ValueError("Could not validate configuration") from e
            self.config = yaml.safe_load(config)
        else:
            self._orig_config = yaml.dump(config)
            self.config = config
        fontc_args.modify_config(self.config)

        self.known_operations = OperationRegistry(use_fontc=fontc_args.use_fontc)
        self.ninja_file_name = fontc_args.build_file_name()
        self.writer = Writer(open(self.ninja_file_name, "w"))
        self.named_files = {}
        self.used_operations = set([])
        self.graph = nx.DiGraph()
        self.recipe = {}  # This will be the filled-in version

        if "recipeProvider" not in self.config and "recipe" not in self.config:
            self.config["recipeProvider"] = "googlefonts"

        if "recipeProvider" in self.config:
            # Store the automatic recipe but allow user-defined steps to override
            automatic_recipe = self.call_recipe_provider()
            assert isinstance(automatic_recipe, dict)
            self.recipe = self.perform_overrides(automatic_recipe)
        elif "recipe" in self.config:
            self.recipe = self.config["recipe"]
        self.validate_recipe()

    def perform_overrides(self, automatic_recipe: Recipe):
        if "recipe" not in self.config:
            return automatic_recipe
        for file, steps in self.config["recipe"].items():
            file = str(file)
            if file in automatic_recipe:
                if all("postprocess" in step for step in steps):
                    # An override step full of postprocesses *adds* to the recipe
                    automatic_recipe[file].extend(steps)
                else:
                    # Something that looks like a recipe overrides
                    automatic_recipe[file] = steps
            else:
                # A new file just gets added
                automatic_recipe[file] = steps
        return automatic_recipe

    # The builder proceeds in four steps. The first step is to prepare
    # the recipe, in YAML-like objects. If there isn't a recipe, we use
    # a recipe provider to create one. Once we have one, we do some basic
    # validation checks on the recipe.

    def call_recipe_provider(self) -> Recipe:
        provider = get_provider(self.config["recipeProvider"])
        return provider(self.config, self).write_recipe()

    def validate_recipe(self):
        for target, steps in self.recipe.items():
            if steps[0].get("source") is None:
                raise ValueError("First step must have a 'source' key")
            seen_postprocess = False
            for step in steps:
                if "operation" in step:
                    if seen_postprocess:
                        raise ValueError(
                            f"Seen operation step '{step['operation']}' after postprocessing step in {target}"
                        )
                elif "postprocess" in step:
                    seen_postprocess = True

    # The next step is to turn the recipe into a set of Python objects;
    # these can store a bit more information than our simple YAML-like
    # data.
    def config_to_objects(self):
        for target, steps in self.recipe.items():
            newsteps = []
            for step in steps:
                if "source" in step:
                    newsteps.append(self.source_step_to_object(step))
                elif "operation" in step:
                    newsteps.append(self.operation_step_to_object(step))
                elif "postprocess" in step:
                    newsteps.append(self.postprocess_step_to_object(step))
                else:
                    raise ValueError(f"Unknown step type: {step}")
            steps[:] = newsteps

    def source_step_to_object(self, step):
        source: str = step["source"]
        return self._ensure_named_file(source, type="source")

    def glyphs_to_ufo(self, source):
        source = Path(source)
        directory = source.resolve().parent
        output = str(Path(directory) / source.with_suffix(".designspace").name)
        glyph_data = self.config.get("glyphData")
        if glyph_data is not None:
            glyph_data = glyph_data
        FontProject().run_from_glyphs(
            str(source.resolve()),
            **{
                "output": ["ufo"],
                "output_dir": directory,
                "master_dir": directory,
                "designspace_path": output,
                "ufo_structure": "json",
                "glyph_data": glyph_data,
            },
        )
        return source.with_suffix(".designspace").name

    def operation_step_to_object(self, step):
        operation = step.get("operation") or step.get("postprocess")
        cls = self.known_operations.get(operation)
        if cls is None:
            raise ValueError(f"Unknown operation {operation}")
        if operation not in self.used_operations:
            self.used_operations.add(operation)
            cls.write_rules(self.writer)
        step = cls(original=step)
        step.convert_dependencies(self)
        return step

    def postprocess_step_to_object(self, step):
        obj = self.operation_step_to_object(step)
        obj.postprocess = True
        return obj

    # The third step is to build a graph and sew together the objects
    # such that the output step of one is fed is the input to the next,
    # and vice versa.
    def build_graph(self):
        Copy().write_rules(self.writer)
        for target, steps in self.recipe.items():
            if target not in self.named_files:
                self._ensure_named_file(target, type="binary")
            for step in steps:
                if isinstance(step, File):
                    self.graph.add_node(step)

        for target, steps in self.recipe.items():
            self._build_graph(target, steps)

    def _ensure_named_file(self, file, type="binary"):
        if file not in self.named_files:
            self.named_files[file] = File(file, type)
            self.graph.add_node(self.named_files[file])
        return self.named_files[file]

    def _build_graph(self, target, steps):
        current = None
        if not isinstance(target, File):
            target = self.named_files[target]
        if target not in self.graph:
            self.graph.add_node(target)
        all_operations = [
            s for s in steps if isinstance(s, OperationBase) and not s.postprocess
        ]
        if not all_operations:
            raise ValueError(f"No operations in {target}")
        last_operation = all_operations[-1]
        binary = None
        for ix, step in enumerate(steps):
            if isinstance(step, File):
                # This is a source: entry, saying that we should either
                # start the build process with a particular source, or
                # switch to a new source midway through.
                if current:
                    if current.path is None:
                        current.path = step.path
                        current = step
                    elif current.path != step.path:
                        # We did something, and now we are switching source.
                        # We assume that this new source has been produced
                        # by the previous operation. So walk backwards,
                        # find that operation, and add this as an out-node.
                        parents = list(self.graph.predecessors(current))
                        if len(parents) != 1:
                            raise ValueError(
                                f"Multiple rules apparently produced {current.path}"
                            )
                        edge = self.graph[parents[0]][current]
                        edge["operation"]._targets = [step]
                        current.path = step.path
                current = step
            else:
                # This is an ordinary operation.
                if current is None:
                    # We need to start somewhere!
                    raise ValueError(f"Operation without source in target {target}")

                # If there is an edge from the source to the operation, then follow it
                # This means that another target has also depended on this operation.
                existing_edge = edge_with_operation(self.graph[current], step)
                # XXX handle postprocessing operations here
                previous = current
                if existing_edge:
                    # print(f"Found an edge from {current.path} via {step.opname}")
                    current = existing_edge
                    # If we are expecting a different target name, copy the
                    # file to rename it.
                    if current.path != target.path and step.object_equals(
                        last_operation
                    ):
                        # print(f"Expected it to be {target.path}, copying")
                        copy_operation = Copy()
                        copy_operation.set_source(current)
                        copy_operation.set_target(target)
                        self.graph.add_edge(current, target, operation=copy_operation)
                        current = target
                else:
                    # We are the first to run this operation, so we need to
                    # create a new edge in the graph.
                    if step.postprocess:
                        # In a post-processing step, we expect:
                        # The target is the stamp file
                        # The source is the terminal binary
                        # The implicit files are any previous stamp files
                        binary = File(step.stamppath)
                        self.graph.add_node(binary)
                        step._sources = last_operation.targets
                        parents = list(self.graph.predecessors(current))
                        previous_edge = self.graph[parents[0]][current]
                        step.implicit = [current]
                        # Check if postprocess args reference other recipe
                        # targets (cross-target dependencies). E.g. gen-stat
                        # may need both the upright and italic variants built.
                        args_str = step.original.get("args", "")
                        for recipe_target in self.recipe:
                            if (
                                recipe_target in args_str
                                and recipe_target != target.path
                            ):
                                dep_file = self.named_files.get(recipe_target)
                                if dep_file and dep_file not in step.implicit:
                                    step.implicit.append(dep_file)
                        self.graph.add_edge(current, binary, operation=step)
                    else:
                        step.set_source(previous)
                        if step.object_equals(last_operation):
                            binary = target
                            binary.terminal = True
                            step.set_target(binary)
                        elif step.targets:  #  Step already knows its own target
                            binary = step.targets[0]
                        else:
                            if self.config.get("logLevel") == "DEBUG":
                                debugnames = []
                                for s in steps[0 : ix + 1]:
                                    if isinstance(s, OperationBase):
                                        debugnames.append(s.opname)
                                    if isinstance(s, File):
                                        debugnames = [s.basename]
                                binary = File(
                                    os.path.join(
                                        tempfile.gettempdir(),
                                        f"builder-{target.basename}-{'-'.join(debugnames)}",
                                    )
                                )
                            else:
                                binary = File(NamedTemporaryFile().name)
                            self.graph.add_node(binary)
                            step.set_target(binary)
                        self.graph.add_edge(current, binary, operation=step)
                        if str(current.path) == str(binary):
                            raise ValueError(
                                f"Adding a circular edge: {current.path}->{step.opname}->{binary}"
                            )
                    # print(
                    #     f"Creating an edge from {current.path} to {binary} via {step.opname}"
                    # )
                    current = binary

    # Finally we walk the graph. We do another validation pass to make
    # sure that the operations make sense, and then we emit the ninja rules.
    def walk_graph(self):
        # A step which consumes a file that other steps postprocess in
        # place must wait for those postprocesses to finish: their stamp
        # files become implicit dependencies of the consuming step.
        for node in self.graph.nodes:
            stamps, consumers = [], []
            for successor in self.graph.successors(node):
                operation = self.graph[node][successor].get("operation")
                if operation is None:
                    continue
                if operation.postprocess:
                    stamps.append(successor)
                else:
                    consumers.append(operation)
            for operation in consumers:
                for stamp in stamps:
                    if stamp not in operation.implicit:
                        operation.implicit.add(stamp)

        actions = defaultdict(list)
        final_targets = []
        for source, target in nx.algorithms.traversal.edge_bfs(self.graph):
            edge = self.graph[source][target]
            if "operation" not in edge:
                continue  # ???
            edge["operation"].validate()
            actions[(source, edge["operation"])].append(target)
            if not list(self.graph.successors(target)):
                final_targets.append(escape_path(target.path))

        for (source, operation), targets in actions.items():
            operation.build(self.writer)

        assert len(final_targets), "No final targets"
        self.writer.default(final_targets)
        self.writer.close()

    def draw_graph(self):
        import pydot

        dot = subprocess.run(
            ["ninja", "-t", "graph", "-f", self.ninja_file_name], capture_output=True
        )
        graphs = pydot.graph_from_dot_data(dot.stdout.decode("utf-8"))
        targets = self.recipe.keys()
        if graphs and graphs[0]:
            for g in graphs[0].get_nodes():
                if g.get_label() and g.get_label().endswith('stamp"'):
                    g.set_style("filled")
                    g.set_fillcolor("#ffcccc")
                    g.set_label("Stamp")
                elif g.get_label() and g.get_label().startswith('"' + gettempdir()):
                    g.set_style("filled")
                    g.set_fillcolor("#ffcccc")
                    g.set_label("Tempfile")
                elif g.get_label() and g.get_label()[1:-1] in targets:
                    g.set_style("filled")
                    g.set_fillcolor("#ccccff")
            graphs[0].write_png("graph.png")
        else:
            print("Could not parse ninja build file")

    def clean(self):
        if hasattr(self, "config") and isinstance(self.config, dict):
            cleanUp = self.config.get("cleanUp")
            if cleanUp == True:
                print("Cleaning up temporary files...")

                for file in [self.ninja_file_name, "./.ninja_log"]:
                    if os.path.exists(file):
                        os.remove(file)

                # in the fontc/crater case it's possible that multiple instances of
                # gftools builder are running at once, in the same directory,
                # in which case we don't want to delete this out from under another process.
                # since we should have deleted our own .ninja file above, if any still
                # exist we will assume another instance of gftools is running.
                if not any(path.endswith("ninja") for path in os.listdir()):
                    if os.path.exists("instance_ufos"):
                        shutil.rmtree("instance_ufos")
                else:
                    print("another .ninja file exists, leaving instance_ufos in place")

                print("Done cleaning up temporary files")
        else:
            print("Configuration not found or invalid, skipping cleanup.")


def main(args=None):
    import argparse

    import yaml

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--graph", help="Draw a graph of the build process", action="store_true"
    )
    parser.add_argument("--no-ninja", help="Do not run ninja", action="store_true")
    parser.add_argument(
        "--generate",
        help="Just generate and output recipe from recipe builder",
        action="store_true",
    )
    parser.add_argument(
        "--experimental-fontc",
        help=f"Use fontc instead of fontmake. Argument is path to the fontc executable",
        type=Path,
    )

    parser.add_argument(
        "--experimental-simple-output",
        help="generate a reduced set of targets, and copy them to the provided directory",
        type=Path,
    )

    parser.add_argument(
        "--experimental-single-source",
        help="only compile the single named source file",
        type=str,
    )

    parser.add_argument("config", help="Path to config file or source file", nargs="+")
    args = parser.parse_args(args)
    fontc_args = FontcArgs(args)
    yaml_files = []
    source_files = []
    for config in args.config:
        if config.endswith(".yaml") or config.endswith(".yml"):
            yaml_files.append(config)
        else:
            source_files.append(config)
    if yaml_files and source_files:
        raise ValueError(
            "Cannot mix YAML config files and font source files on command line"
        )
    if source_files:
        config = {
            "sources": source_files,
            "outputDir": "fonts/",
        }
    else:
        if len(args.config) > 1:
            raise ValueError("Only one config file can be given for now")
        config = args.config[0]

    pd = GFBuilder(config, fontc_args=fontc_args)
    if args.generate:
        config = pd.config
        config["recipe"] = pd.recipe
        print(yaml.dump(config))
        return
    pd.config_to_objects()
    pd.build_graph()
    pd.walk_graph()
    if args.graph:
        pd.draw_graph()
    if not args.no_ninja:
        atexit.register(pd.clean)
        raise SystemExit(_program("ninja", ["-f", pd.ninja_file_name]))



#!/usr/bin/env python3
import os
import argparse
from fontTools.ttLib import TTFont


def check_glyph_in_font(font_path, char):
    """Return True if the Unicode char exists in the given font file."""
    try:
        font = TTFont(font_path)
        cmap = font.getBestCmap()  # maps Unicode codepoints -> glyph names
        codepoint = ord(char)
        return codepoint in cmap
    except Exception as e:
        print(f"[ERROR] Could not open {font_path}: {e}")
        return False


def scan_directory(root_dir, char):
    """Walk through directories and check each .ttf file for the character."""
    for dirpath, _, filenames in os.walk(root_dir):
        for filename in filenames:
            if filename.lower().endswith(".ttf"):
                font_path = os.path.join(dirpath, filename)
                if check_glyph_in_font(font_path, char):
                    print(f"[FOUND] '{char}' (U+{ord(char):04X}) in {font_path}")


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Check if TTF fonts in a directory contain a specific Unicode glyph."
    )
    parser.add_argument("directory", help="Root directory to scan for TTF files")
    parser.add_argument(
        "glyph", help="The actual character to search for (e.g., 'A', '€', '你')"
    )
    args = parser.parse_args(args)

    if len(args.glyph) != 1:
        parser.error("Please provide exactly one character as the glyph argument.")

    scan_directory(args.directory, args.glyph)


if __name__ == "__main__":
    main()



import os

from gftools.builder.operations import OperationBase


class BuildFvarInstances(OperationBase):
    description = "Run gftools-gen-fvar-instances"
    rule = "gftools-gen-fvar-instances --inplace $in $args"



"""CoreText rendering backend (macOS).

Uses CoreText for shaping and CGBitmapContext for rasterisation. Runs only
on macOS — pyobjc-framework-CoreText / pyobjc-framework-Quartz must be
installed (declared as platform-conditional deps in pyproject.toml).
"""

from __future__ import annotations

from pathlib import Path

from PIL import Image

try:
    import CoreText
    import Quartz
    from Foundation import NSData
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "gftools was installed without the QA dependencies. To install the "
        "dependencies, see the ReadMe, "
        "https://github.com/googlefonts/gftools#installation"
    )

PADDING = 2


def render_row(
    font_path: Path,
    text: str,
    ppem: int,
    variations: dict[str, float] | None = None,
    *,
    target_height: int,
    baseline_y: int,
) -> Image.Image:
    ct_font = _make_ct_font(font_path, ppem, variations)
    line = _make_line(ct_font, text)

    width_d, _ascent, _descent, _leading = _measure(line)
    width = max(int(width_d) + PADDING * 2, 1)
    # CG's drawing coords are bottom-left origin; convert baseline-from-top.
    baseline_y_cg = target_height - baseline_y

    ctx = _gray_bitmap_context(width, target_height)
    Quartz.CGContextSetGrayFillColor(ctx, 1.0, 1.0)
    Quartz.CGContextFillRect(ctx, ((0, 0), (width, target_height)))
    Quartz.CGContextSetGrayFillColor(ctx, 0.0, 1.0)
    Quartz.CGContextSetTextPosition(ctx, PADDING, baseline_y_cg)
    CoreText.CTLineDraw(line, ctx)

    pixels = Quartz.CGBitmapContextGetData(ctx).as_buffer(width * target_height)
    img = Image.frombytes("L", (width, target_height), bytes(pixels))
    return img.convert("RGB")


def _make_ct_font(font_path: Path, ppem: int, variations: dict[str, float] | None):
    data = NSData.dataWithContentsOfFile_(str(font_path))
    if data is None:
        raise FileNotFoundError(font_path)
    provider = Quartz.CGDataProviderCreateWithCFData(data)
    cg_font = Quartz.CGFontCreateWithDataProvider(provider)
    if cg_font is None:
        raise RuntimeError(f"CoreText could not load font: {font_path}")
    ct_font = CoreText.CTFontCreateWithGraphicsFont(cg_font, ppem, None, None)
    if not variations:
        return ct_font
    return _apply_variations(ct_font, variations, ppem)


def _apply_variations(ct_font, variations: dict[str, float], ppem: int):
    var_attr = {_tag_to_int(tag): float(value) for tag, value in variations.items()}
    descriptor = CoreText.CTFontCopyFontDescriptor(ct_font)
    new_desc = CoreText.CTFontDescriptorCreateCopyWithAttributes(
        descriptor, {CoreText.kCTFontVariationAttribute: var_attr}
    )
    return CoreText.CTFontCreateWithFontDescriptor(new_desc, ppem, None)


def _make_line(ct_font, text: str):
    attrs = {CoreText.kCTFontAttributeName: ct_font}
    astr = CoreText.CFAttributedStringCreate(None, text, attrs)
    return CoreText.CTLineCreateWithAttributedString(astr)


def _measure(line) -> tuple[float, float, float, float]:
    ascent, descent, leading = 0.0, 0.0, 0.0
    width = CoreText.CTLineGetTypographicBounds(line, None, None, None)
    if isinstance(width, tuple):
        width, ascent, descent, leading = width
    return width, ascent, descent, leading


def _gray_bitmap_context(width: int, height: int):
    color_space = Quartz.CGColorSpaceCreateDeviceGray()
    return Quartz.CGBitmapContextCreate(
        None, width, height, 8, width, color_space, Quartz.kCGImageAlphaNone
    )


def _tag_to_int(tag: str) -> int:
    if len(tag) != 4:
        raise ValueError(f"axis tag must be 4 chars: {tag!r}")
    b = tag.encode("ascii")
    return (b[0] << 24) | (b[1] << 16) | (b[2] << 8) | b[3]



#!/usr/bin/env python3
"""Compare metadata from different servers and generate a vimdiff HTML report."""
import json
import subprocess
import os
from glob import glob
import tempfile
import requests
from gftools.push.servers import (
    SANDBOX_META_URL,
    PRODUCTION_META_URL,
    SANDBOX_VERSIONS_URL,
    PRODUCTION_VERSIONS_URL,
    server_auth,
)
import argparse
import sys

SANDBOX_AUTH = server_auth("sandbox")
PRODUCTION_AUTH = server_auth("production")


def munge_meta(obj):
    family_res = {}
    for family in obj["familyMetadataList"]:
        family_res[family["family"]] = {
            k: v
            for k, v in family.items()
            if k
            not in [
                "lastModified",
                "popularity",
                "trending",
                "defaultSort",
                "size",
                "dateAdded",
                #            "subsets"
            ]
        }
    obj["familyMetadataList"] = family_res
    obj["axisRegistry"].sort(key=lambda x: x["tag"])
    return obj


def munge_fontv(obj):
    return obj


def munge_designer(obj):
    return obj


def munge_family(obj):
    if "family" not in obj:
        return "Not in server yet"
    obj["coverage"] = list(obj["coverage"].keys())
    obj.pop("stats")
    obj.pop("size")
    obj.pop("lastModified")
    return obj


def pr_json(number):
    cmd = ["gh", "pr", "view", number, "--json", "title,labels,number,files"]
    msg = subprocess.run(cmd, check=True, capture_output=True)
    data = json.loads(msg.stdout.decode("utf-8"))
    return data


def levenstein(a, b):
    if len(a) < len(b):
        return levenstein(b, a)

    if len(b) == 0:
        return len(a)

    if len(b) == 1:
        return 1 if a.find(b) != -1 else len(a)

    # let a and b be lists of characters
    a = list(a)
    b = list(b)
    alen = len(a)
    blen = len(b)
    current_row = range(alen + 1)

    for i in range(1, blen + 1):
        previous_row, current_row = current_row, [i] * (alen + 1)
        for j in range(1, alen + 1):
            insertions = previous_row[j] + 1
            deletions = current_row[j - 1] + 1
            substitutions = previous_row[j - 1] + (a[j - 1] != b[i - 1])
            current_row[j] = min(insertions, deletions, substitutions)

    return current_row[alen]


def doc_in_server(doc, sb_data, prod_data):
    sb_res = levenstein(json.dumps(sb_data), doc)
    prod_res = levenstein(json.dumps(prod_data), doc)

    best = min(sb_res, prod_res)
    if sb_res == best:
        return "sb"
    elif prod_res == best:
        return "prod"


def which_server_metadata(string, family):
    munge = munge_family
    sb_meta = munge(
        json.loads(
            requests.get(f"{SANDBOX_META_URL}/{family}", auth=SANDBOX_AUTH).text[4:]
        )
    )
    prod_meta = munge(
        json.loads(
            requests.get(f"{PRODUCTION_META_URL}/{family}", auth=PRODUCTION_AUTH).text[
                4:
            ]
        )
    )
    res = doc_in_server(string, sb_meta, prod_meta)


def pr_type(data):
    if data["labels"][0]["name"] == "I Metadata/OFL":
        file_to_check = next(f for f in data["files"] if "METADATA.pb" in f["path"])
        if file_to_check:
            with open(file_to_check["path"], "r") as doc:
                which_server_metadata(doc.read(), "Huninn")


def get_designer(url, designer, auth=None):
    meta = requests.get(url, auth=auth).json()
    for family in meta["familyMetadataList"]:
        for family_designer in family["designers"]:
            if family_designer == designer:
                family_to_check = family["family"]
                data = json.loads(
                    requests.get(f"{url}/{family_to_check}", auth=auth).text[4:]
                )
                return next(d for d in data["designers"] if d["name"] == designer)
    return {}


def generate_vimdiff_html(sb_file, prod_file, output_file):
    """
    Use vimdiff to compare sb_meta and prod_meta and save the results as an HTML file.
    Automatically handle swap files by deleting them if they exist.
    """
    # Check and delete swap files if they exist
    swap_files = glob(".*.swp")
    for swap_file in swap_files:
        os.remove(swap_file)
    try:
        # Run vimdiff in batch mode and output the results to an HTML file
        cmd = [
            "/usr/bin/vim",
            "-i",
            "NONE",
            "-n",
            "-N",
            "-d",
            sb_file,
            prod_file,
            "-c",
            "windo set wrap",
            "-c",
            "TOhtml",
            "-c",
            f"sav! {output_file}",
            "-c",
            "qall!",
        ]
        subprocess.run(cmd, check=True)
        print(f"Comparison saved to {output_file}")
    except subprocess.CalledProcessError as e:
        print(f"Error running vimdiff: {e}")


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Compare metadata from different servers."
    )
    parser.add_argument("--gf-path", help="path to google/fonts")
    diff_type = parser.add_mutually_exclusive_group()
    diff_type.add_argument("--pr", help="diff a pr")
    diff_type.add_argument("--meta", action="store_true", help="compare font metadata")
    diff_type.add_argument("--fontv", action="store_true", help="compare font version")
    diff_type.add_argument("--family", help="family to compare")
    diff_type.add_argument("--designer", help="designer to compare")

    parser.add_argument("-o", "--out", help="output path to html file")
    args = parser.parse_args(args)
    # TODO make this a required arg if a pr is given
    if args.gf_path:
        os.chdir(args.gf_path)
    if args.pr:
        raise (NotImplementedError("PR diff not implemented yet"))
        data = pr_json(args.pr)
        pr_type(data)
        import pdb

        pdb.set_trace()
        return
    if args.meta:
        munge = munge_meta
        sb_meta = munge(requests.get(SANDBOX_META_URL, auth=SANDBOX_AUTH).json())
        prod_meta = munge(
            requests.get(PRODUCTION_META_URL, auth=PRODUCTION_AUTH).json()
        )
    elif args.fontv:
        munge = munge_fontv
        sb_meta = munge(
            json.loads(requests.get(SANDBOX_VERSIONS_URL, auth=SANDBOX_AUTH).text[4:])
        )
        prod_meta = munge(
            json.loads(
                requests.get(PRODUCTION_VERSIONS_URL, auth=PRODUCTION_AUTH).text[4:]
            )
        )
    elif args.family:
        munge = munge_family
        sb_meta = munge(
            json.loads(
                requests.get(
                    f"{SANDBOX_META_URL}/{args.family}", auth=SANDBOX_AUTH
                ).text[4:]
            )
        )
        prod_meta = munge(
            json.loads(
                requests.get(
                    f"{PRODUCTION_META_URL}/{args.family}", auth=PRODUCTION_AUTH
                ).text[4:]
            )
        )
    elif args.designer:
        munge = munge_designer
        sb_meta = get_designer(SANDBOX_META_URL, args.designer, auth=SANDBOX_AUTH)
        prod_meta = get_designer(
            PRODUCTION_META_URL, args.designer, auth=PRODUCTION_AUTH
        )

    with tempfile.TemporaryDirectory() as temp_dir:
        sb_file = os.path.join(temp_dir, "sb_meta.json")
        prod_file = os.path.join(temp_dir, "prod_meta.json")
        with open(sb_file, "w") as f:
            json.dump(sb_meta, f, indent=4)
        with open(prod_file, "w") as f:
            json.dump(prod_meta, f, indent=4)
        if args.out:
            generate_vimdiff_html(sb_file, prod_file, args.out)
        else:
            generate_vimdiff_html(
                sb_file, prod_file, os.path.join(temp_dir, "diff.html")
            )
            if sys.platform == "linux":
                open_cmd = "xdg-open"
            elif sys.platform == "darwin":
                open_cmd = "open"
            elif sys.platform == "win32":
                open_cmd = "start"
            else:
                raise NotImplementedError("Unsupported OS")
            subprocess.run([open_cmd, os.path.join(temp_dir, "diff.html")])
            print("Hit any key to exit")
            input()


if __name__ == "__main__":
    main()
